// PreviewDialog.cpp : implementation file
//

#include "stdafx.h"
#include "WinPVR.h"
#include "pvrfile.h"
#include "mpeg.h"
#include "general.h"
#include <cmath>
#include <iomanip>
#include <iostream>
#include "PreviewDialog.h"

const int nSkipsizes [3][4] = {15000, 30000, 60000, 120000,
                                2000,  4500,  9000,  18000,
                                 100,   300,   500,   1000};
// CPreviewDialog dialog

IMPLEMENT_DYNAMIC(CPreviewDialog, CDialog)
CPreviewDialog::CPreviewDialog(CWnd* pParent /*=NULL*/)
	: CDialog(CPreviewDialog::IDD, pParent)
  , m_sRuntime(_T(""))
  , m_sPosition(_T("0:00:00.00"))
  , m_i64BeginPos(0)
  , m_bWheelTimer(false)
  , m_nWheelCount (0)
  , m_nNumberOfJobs(0)
  , m_sEndTime(_T(""))
  , m_sBeginTime(_T(""))
  , m_sClipName(_T(""))
  , m_bRipStreamRaw(FALSE)
  , m_bRipAudioRaw(FALSE)
  , m_dwBeginTime(0)
  , m_dwEndTime(0)
  , m_dGotoTime(0)
  , m_oleRealTime(COleDateTime::GetCurrentTime())
{
  m_hBitmap = NULL;
}

CPreviewDialog::~CPreviewDialog()
{
    if (m_hBitmap) 
    {
      DeleteObject(m_hBitmap);
    }
}

void CPreviewDialog::DoDataExchange(CDataExchange* pDX)
{
  CDialog::DoDataExchange(pDX);
  DDX_Control(pDX, IDC_SLIDER_TIME, m_wndSliderTime);
  DDX_Text(pDX, IDC_STATIC_RUNTIME, m_sRuntime);
  DDV_MaxChars(pDX, m_sRuntime, 20);
  DDX_Text(pDX, IDC_STATIC_POSITION, m_sPosition);
  DDX_Text(pDX, IDC_STATIC_JOBS, m_nNumberOfJobs);
  DDX_Text(pDX, IDC_STATIC_POSITION_END, m_sEndTime);
  DDX_Text(pDX, IDC_STATIC_POSITION_BEGIN, m_sBeginTime);
  DDX_Text(pDX, IDC_EDIT_NAME, m_sClipName);
  DDX_Check(pDX, IDC_CHECK_STREAM_RAW, m_bRipStreamRaw);
  DDX_Check(pDX, IDC_CHECK_AUDIO_RAW, m_bRipAudioRaw);
  DDX_Control(pDX, IDC_CHECK_MAGNIFY, m_wndMagnify);
  DDX_Text(pDX, IDC_EDIT_GOTO, m_dGotoTime);
  DDX_Text(pDX, IDC_STATIC_POSITION_REALTIME, m_oleRealTime);
}


BEGIN_MESSAGE_MAP(CPreviewDialog, CDialog)
  ON_NOTIFY(NM_RELEASEDCAPTURE, IDC_SLIDER_TIME, OnNMReleasedcaptureSliderTime)
  ON_BN_CLICKED(IDC_BUTTON_30SEK_BACK, OnBnClickedButton30sekBack)
  ON_BN_CLICKED(IDC_BUTTON_5SEC_BACK, OnBnClickedButton5secBack)
  ON_BN_CLICKED(IDC_BUTTON_1SEC_BACK, OnBnClickedButton1secBack)
  ON_BN_CLICKED(IDC_BUTTON_1SEC_FOR, OnBnClickedButton1secFor)
  ON_BN_CLICKED(IDC_BUTTON_5SEC_FOR, OnBnClickedButton5secFor)
  ON_BN_CLICKED(IDC_BUTTON_30SEC_FOR, OnBnClickedButton30secFor)
  ON_WM_MOUSEWHEEL()
  ON_WM_TIMER()
  ON_BN_CLICKED(IDC_BUTTON_SET_BEGIN, OnBnClickedButtonSetBegin)
  ON_BN_CLICKED(IDC_BUTTON_SET_END, OnBnClickedButtonSetEnd)
  ON_BN_CLICKED(IDC_BUTTON_ADD_JOB, OnBnClickedButtonAddJob)
  ON_BN_CLICKED(IDC_BUTTON_5MIN_BACK, OnBnClickedButton5minBack)
  ON_BN_CLICKED(IDC_BUTTON_5MIN_FOR, OnBnClickedButton5minFor)
  ON_WM_HSCROLL()
//  ON_WM_KEYDOWN()
ON_WM_LBUTTONDOWN()
//ON_WM_KEYDOWN()
ON_BN_CLICKED(IDC_CHECK_MAGNIFY, OnBnClickedCheckMagnify)
ON_WM_RBUTTONDOWN()
ON_BN_CLICKED(IDC_BUTTON_GOTO, OnBnClickedButtonGoto)
ON_BN_CLICKED(IDOK, OnBnClickedOk)
END_MESSAGE_MAP()


// CPreviewDialog message handlers
bool CPreviewDialog::Init(CPVRFile * pPVRFile)
{
  m_pPVRFile = pPVRFile;
  return true;
}

BOOL CPreviewDialog::OnInitDialog(void)
{
  bool bCancel = false;
  DWORD dwLength;
  CString s;
	CDialog::OnInitDialog();
	// Cancelzeiger des PVRFile Objekts erstellen
  if (m_pPVRFile)
  {
    m_nMagnify = 2;
    for (int i = IDC_BUTTON_5MIN_BACK; i <= IDC_BUTTON_5MIN_FOR; i++)
    {
      s.LoadString (i+10*m_nMagnify);
      GetDlgItem (i)->SetWindowText (s);
    }
    s.LoadString (IDS_STRING_MAGBUTTON + m_nMagnify);
    m_wndMagnify.SetWindowText (s);
    m_sFD = m_pPVRFile->GetFileDescription ();
    m_pPVRFile->SetFilename ("PRV", false);
	  m_pPVRFile->m_pbCancel = &bCancel;
//    m_pPVRFile->GetClusterPTS ();
	  // Auf geht's
    // Create a child bitmap static control.
    m_wndPreview.Create (_T("my static"), WS_CHILD|WS_VISIBLE|SS_BITMAP|SS_CENTERIMAGE, 
      CRect(120,20,520,320), this);
	  m_pPVRFile->RipStreams (0, 0, false);
    m_i64BeginPos = m_pPVRFile->GetCurrentPos ();
    dwLength = (m_sFD->iLengthHours * 3600 + m_sFD->iLengthMinutes * 60 + m_sFD->iLengthSeconds) * 100;
    m_sRuntime.Format ("%d:%02d:%02d",  m_sFD->iLengthHours,  m_sFD->iLengthMinutes, m_sFD->iLengthSeconds);

    
    DecodeSnapshot ("PRV.M2V", "PRV.BMP");
    SetWindowText (m_sFD->cClipName);
    m_sClipName = m_sFD->cClipName;
    if (m_hBitmap) 
    {
      DeleteObject(m_hBitmap);
      m_hBitmap = NULL;
    }
    m_hBitmap = (HBITMAP) LoadImage(NULL, "PRV.BMP", IMAGE_BITMAP,
      400, (int) (400.0 / m_pPVRFile->m_dAspectRatio), LR_LOADFROMFILE);
    m_wndPreview.SetBitmap(m_hBitmap);
    try  
    {   CFile::Remove("PRV.M2V");   CFile::Remove("PRV.BMP");  }
    catch(...)  {  }
    // Date Time:
    m_oleRealTime.SetDateTime 
      (m_sFD->iRecordYear, m_sFD->iRecordMonth, m_sFD->iRecordDay,
      m_sFD->iRecordHour, m_sFD->iRecordMinute, m_sFD->iRecordSecond);
  }
  m_wndSliderTime.SetRange (0, dwLength-200);
  m_wndSliderTime.SetTicFreq (60000);

  UpdateData (FALSE);
  return 0;
}

void CPreviewDialog::OnNMReleasedcaptureSliderTime(NMHDR *pNMHDR, LRESULT *pResult)
{
  bool bCancel = false;

  double dPos;
  DWORD dwPos;
  COleDateTimeSpan oleTimeSpan;
  UpdateData ();
  dPos = m_wndSliderTime.GetPos () / 100;
  if (m_pPVRFile)
  {
    m_pPVRFile->SetFilename ("PRV", false);
	  m_pPVRFile->m_pbCancel = &bCancel;
	  // Auf geht's
	  m_pPVRFile->RipStreams (dPos, 0, false);
    m_pPVRFile->m_pInfoVideo->m_bPatchSeqHeader = false;
    m_i64CurPos = m_pPVRFile->GetCurrentPos ();
    DecodeSnapshot ("PRV.M2V", "PRV.BMP");
    dwPos = (DWORD) ((m_i64CurPos - m_i64BeginPos) / 900);
    m_sPosition.Format ("%d:%02d:%02d.%02d", (dwPos / 360000), (dwPos / 6000) % 60, (dwPos / 100) % 60, dwPos % 100);
    if (m_hBitmap) 
    {
      DeleteObject(m_hBitmap);
      m_hBitmap = NULL;
    }
    m_hBitmap = (HBITMAP) LoadImage(NULL, "PRV.BMP", IMAGE_BITMAP,
      400, (int) (400.0 / m_pPVRFile->m_dAspectRatio), LR_LOADFROMFILE);
    m_wndPreview.SetBitmap(m_hBitmap);
    try  
    {   CFile::Remove("PRV.M2V");   CFile::Remove("PRV.BMP");  }
    catch(...)  {  }
  }
  if (pResult)
    *pResult = 0;
  m_oleRealTime.SetDateTime 
    (m_sFD->iRecordYear, m_sFD->iRecordMonth, m_sFD->iRecordDay,
    m_sFD->iRecordHour, m_sFD->iRecordMinute, m_sFD->iRecordSecond);
  oleTimeSpan.SetDateTimeSpan (0, 0, 0,  (int) ((m_i64CurPos - m_i64BeginPos) / 90000));
  m_oleRealTime += oleTimeSpan;
  UpdateData (FALSE);
}


void CPreviewDialog::OnBnClickedButton5minBack()
{
  OnBnClickedButtonX (-1, 3);
}
void CPreviewDialog::OnBnClickedButton30sekBack()
{
  OnBnClickedButtonX (-1, 2);
}
void CPreviewDialog::OnBnClickedButton5secBack()
{
  OnBnClickedButtonX (-1, 1);
}
void CPreviewDialog::OnBnClickedButton1secBack()
{
  OnBnClickedButtonX (-1, 0);
}

void CPreviewDialog::OnBnClickedButton1secFor()
{
  OnBnClickedButtonX (1, 0);
}
void CPreviewDialog::OnBnClickedButton5secFor()
{
  OnBnClickedButtonX (1, 1);
}
void CPreviewDialog::OnBnClickedButton30secFor()
{
  OnBnClickedButtonX (1, 2);
}
void CPreviewDialog::OnBnClickedButton5minFor()
{
  OnBnClickedButtonX (1, 3);
}
int CPreviewDialog::OnBnClickedButtonX(int nFactor, int nIndex)
{
  int nPos = m_wndSliderTime.GetPos ();
  nPos += nFactor * nSkipsizes[m_nMagnify][nIndex];
  if (nPos > m_wndSliderTime.GetRangeMax ()) nPos = m_wndSliderTime.GetRangeMax ();
  if (nPos < 0) nPos = 0;
  m_wndSliderTime.SetPos (nPos);
  OnNMReleasedcaptureSliderTime (NULL, NULL);
  return 0;
}

BOOL CPreviewDialog::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
  KillTimer (1);
  SetTimer (1, 500, NULL);
  m_nWheelCount += zDelta;
  m_bWheelTimer = true;
  int nPos = m_wndSliderTime.GetPos ();
  nPos += m_nWheelCount;
  if (nPos > m_wndSliderTime.GetRangeMax ()) nPos = m_wndSliderTime.GetRangeMax ();
  if (nPos < 0) nPos = 0;
  m_sPosition.Format ("%d:%02d:%02d.%02d", (nPos / 360000), (nPos / 6000) % 60, (nPos / 100) % 60, nPos % 100);
  UpdateData (FALSE);

  return CDialog::OnMouseWheel(nFlags, zDelta, pt);
}

void CPreviewDialog::OnTimer(UINT nIDEvent)
{
  if (nIDEvent == 1)
  {
    if (m_bWheelTimer)
    {
      KillTimer (1);
      SetTimer (1, 250, NULL);
      m_bWheelTimer = false;
    }
    else
    {
      int nPos = m_wndSliderTime.GetPos ();
      nPos += m_nWheelCount;
      if (nPos > m_wndSliderTime.GetRangeMax ()) nPos = m_wndSliderTime.GetRangeMax ();
      if (nPos < 0) nPos = 0;
      m_wndSliderTime.SetPos (nPos);
      OnNMReleasedcaptureSliderTime (NULL, NULL);
      m_nWheelCount = 0;
      KillTimer (1);
    }
  }
  CDialog::OnTimer(nIDEvent);
}

void CPreviewDialog::OnBnClickedButtonSetBegin()
{
  UpdateData ();
  CWnd * wnd;
  m_sBeginTime = m_sPosition;
  m_dwBeginTime = m_wndSliderTime.GetPos ();
  if (!m_sEndTime.IsEmpty () && (m_dwBeginTime < m_dwEndTime))
  {
    m_wndSliderTime.SetSelection (m_dwBeginTime, m_dwEndTime);
  }
  else
  {
    m_wndSliderTime.SetSelection (m_dwBeginTime, m_dwBeginTime);
  }
  UpdateData (FALSE);
  wnd = m_wndSliderTime.SetFocus ();
  wnd->SetFocus ();
}

void CPreviewDialog::OnBnClickedButtonSetEnd()
{
  UpdateData ();
  CWnd * wnd;
  m_dwEndTime = m_wndSliderTime.GetPos ();
  if (!m_sBeginTime.IsEmpty () && (m_dwBeginTime < m_dwEndTime))
  {
    m_wndSliderTime.SetSelection (m_dwBeginTime, m_dwEndTime);
  }
  else
  {
    m_wndSliderTime.SetSelection (m_dwEndTime, m_dwEndTime);
  }
  m_sEndTime = m_sPosition;
  UpdateData (FALSE);
  wnd = m_wndSliderTime.SetFocus ();
  wnd->SetFocus ();
}

void CPreviewDialog::OnBnClickedButtonAddJob()
{
 	DWORD dwStunde, dwMinute, dwSecs , dwMS;
  double dSecs;
	UpdateData ();
	// Falls in der Form h:mm:ss eingegeben
	sscanf (m_sBeginTime, "%d:%02d:%02d.%02d", &dwStunde, &dwMinute, &dwSecs, &dwMS);
  dSecs = dwSecs + (double) dwMS / 100;
	m_structJobs[m_nNumberOfJobs].dwOffset = dwStunde * 3600 + dwMinute * 60 + dSecs;
	sscanf (m_sEndTime, "%d:%02d:%02d.%02d", &dwStunde, &dwMinute, &dwSecs, &dwMS);
  dSecs = dwSecs + (double) dwMS / 100;
	m_structJobs[m_nNumberOfJobs].dwRipzeit = dwStunde * 3600 + dwMinute * 60 + dSecs;
  if (m_structJobs[m_nNumberOfJobs].dwRipzeit < m_structJobs[m_nNumberOfJobs].dwOffset) return;
  m_structJobs[m_nNumberOfJobs].dwRipzeit -= m_structJobs[m_nNumberOfJobs].dwOffset;
  m_structJobs[m_nNumberOfJobs].bMakeBackup = false;
  m_structJobs[m_nNumberOfJobs].bRestoreBackup = false;
  m_structJobs[m_nNumberOfJobs].bRipAudio = true;
  m_structJobs[m_nNumberOfJobs].bRipVideo = true;
  m_structJobs[m_nNumberOfJobs].bRipAudioRaw = m_bRipAudioRaw;
  m_structJobs[m_nNumberOfJobs].bRipVideoRaw = m_bRipStreamRaw;
  m_structJobs[m_nNumberOfJobs].bSplitRecording = false;
  m_structJobs[m_nNumberOfJobs].nIndex = m_pPVRFile->m_nFileNo;
  m_structJobs[m_nNumberOfJobs].sFilename = m_sClipName;
  m_structJobs[m_nNumberOfJobs].fSizeInGB =  ((double)(m_sFD->iAudioClusterCount + m_sFD->iVideoClusterCount) / 10.0) *
    (double) m_structJobs[m_nNumberOfJobs].dwRipzeit / (double) (m_sFD->iLengthHours * 3600 + m_sFD->iLengthMinutes * 60 + m_sFD->iLengthSeconds);
  m_nNumberOfJobs++;
  UpdateData (FALSE);
}


void CPreviewDialog::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  DWORD dwPos;
  COleDateTimeSpan oleTimeSpan;
  // TODO: Add your message handler code here and/or call default
  if (nSBCode == TB_THUMBTRACK)
  {
    UpdateData (TRUE);
    dwPos = ((CSliderCtrl *) pScrollBar)->GetPos ();
    m_sPosition.Format ("%d:%02d:%02d.%02d", (dwPos / 360000), (dwPos / 6000) % 60, (dwPos / 100) % 60, dwPos % 100);
    m_oleRealTime.SetDateTime 
      (m_sFD->iRecordYear, m_sFD->iRecordMonth, m_sFD->iRecordDay,
      m_sFD->iRecordHour, m_sFD->iRecordMinute, m_sFD->iRecordSecond);
    oleTimeSpan.SetDateTimeSpan (0, 0, 0,  (int) (dwPos / 100));
    m_oleRealTime += oleTimeSpan;
    UpdateData (FALSE);
  }
  CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
}

void CPreviewDialog::OnLButtonDown(UINT nFlags, CPoint point)
{
  // TODO: Add your message handler code here and/or call default
  SetFocus ();
  CDialog::OnLButtonDown(nFlags, point);
}


void CPreviewDialog::OnBnClickedCheckMagnify()
{
  CString s;
  UpdateData ();
  m_nMagnify = m_wndMagnify.GetCheck ();
  s.LoadString (IDS_STRING_MAGBUTTON + m_nMagnify);
  m_wndMagnify.SetWindowText (s);
  for (int i = IDC_BUTTON_5MIN_BACK; i <= IDC_BUTTON_5MIN_FOR; i++)
  {
    s.LoadString (i+10*m_nMagnify);
    GetDlgItem (i)->SetWindowText (s);
  }
}


void CPreviewDialog::OnRButtonDown(UINT nFlags, CPoint point)
{
  m_nMagnify = (m_nMagnify + 1) % 3;
  m_wndMagnify.SetCheck (m_nMagnify);
  OnBnClickedCheckMagnify ();
  CDialog::OnRButtonDown(nFlags, point);
}

void CPreviewDialog::OnBnClickedButtonGoto()
{
  int nPos;
  UpdateData ();
  nPos =  ((DWORD) m_dGotoTime) % 100; // Sekunden
  nPos += ((DWORD) (m_dGotoTime / 100) % 100) * 60; // Minuten
  nPos += (DWORD) (m_dGotoTime / 10000) * 3600; // Stunden
  nPos *= 100;
  nPos += (DWORD) (m_dGotoTime * 100) % 100;
  if (nPos > m_wndSliderTime.GetRangeMax ()) nPos = m_wndSliderTime.GetRangeMax ();
  if (nPos < 0) nPos = 0;
  m_wndSliderTime.SetPos (nPos);
  OnNMReleasedcaptureSliderTime (NULL, NULL);
}

void CPreviewDialog::OnBnClickedOk()
{
  CWnd * pWnd;
  if (GetFocus () == (pWnd = GetDlgItem (IDC_EDIT_GOTO)))
  {
    ((CEdit *)pWnd)->SetSel (0, -1);
    OnBnClickedButtonGoto ();
    return;
  }
  // TODO: Add your control notification handler code here
  OnOK();
}
