#include "stdafx.h"
#include "CRawFile.h"
#include "mpeg.h"
#include <cmath>
#include <iomanip>
#include <iostream>
using namespace std;

CRawFile::CRawFile(const string &Drivename, unsigned int FileNo)
{
	HANDLE			hFile;
	DWORD			dwBytesRead;

	*pfsLog << "CRawFile::CRawFile(char *Drivename, int FileNo)" << endl;

	m_sDrivename=Drivename;

	if (OpenDrive(hFile, Drivename))
	{
		SetToPhysicalSector(hFile, TOC_OFFSET_SECTORS + FileNo);
		ReadFile(hFile, &m_sfdFile,sizeof(m_sfdFile),&dwBytesRead,NULL);
	}
	else
	{
		*pfsLog << "could not open drive " << Drivename << " for file " << FileNo << endl;
	}
	CloseHandle(hFile);

	m_RecordNoOkay=(FileNo==m_sfdFile.iRecordNo);
	
}

CRawFile::~CRawFile()
{
	*pfsLog << "CRawFile::~CRawFile() for " << m_sfdFile.iRecordNo << endl;
}

bool CRawFile::Deleted()
{
	return !m_RecordNoOkay;
}

string CRawFile::Filename()
{
	char	buf[20];
	string	Filename="Out";
	ltoa((long)m_sfdFile.iRecordNo,buf,10);
	Filename.append(buf);
	return Filename;
}



bool CRawFile::GetStreamIDs(CBits &InStream, unsigned int &StreamID)
{
	int					i;
	unsigned __int64	ui64BitPointer;
	unsigned __int32	ui32Next4Bytes;

	StreamID=0;
	*pfsLog << "GetStreamIDs" << endl;

	*pfsLog << "try to rewind";
	if (!InStream.Rewind())
	{
		*pfsLog << "... failed..." << endl;
		return false;
	}
	*pfsLog << "... okay" << endl;

	ui64BitPointer=0;
	while (InStream.NextStartCode(MBYTE * 10 * 10) && ui64BitPointer < (__int64) MBYTE * 80 * 10)
	{
		ui32Next4Bytes=InStream.Next4Bytes();
		//ui64BitPointer=InStream.NextBits(32);
		for (i=0xBD; i < 0xF0;i++)
		{
			switch(GetStreamType(i))
			{
			case OTHER:
				break;
			default:
				if (IsPES_Packet(InStream,i))
				{
					StreamID=i;
					return true;
				}
				break;
			}
		}
		InStream.MoveBytes(1);
		InStream.BitPointer(ui64BitPointer);
	}

	return false;
}


bool CRawFile::GetFirstPTS(CBits* pcbStream, unsigned int uiStreamID, __int64 &i64PTS)
{
	unsigned __int32	ui32StartCode = 0x100 + uiStreamID;
	unsigned __int32	ui32Next4Bytes;
	// PES Packet
	CPES_Packet*		pppPES=NULL;
	// Streamtype
	STREAM_TYPE			uiStreamType=GetStreamType(uiStreamID);

	i64PTS=0;

	*pfsLog << setiosflags(ios::fixed) << resetiosflags(ios::dec) << setiosflags(ios::hex); 
	*pfsLog << endl << "FirstPTS " ;

	switch (uiStreamType)
	{
	case VIDEO:
		*pfsLog << "VIDEO ";
		break;
	case AUDIO:
		*pfsLog << "AUDIO ";
		break;
	case PRIVATE_1:
		*pfsLog << "PRIVATE_1 " ;
		break;
	case PRIVATE_2:
		*pfsLog << "PRIVATE_2 " ;
		break;
	}
	*pfsLog << (int)ui32StartCode << endl;


	i64PTS=0;
	*pfsLog << "try to rewind";
	if (!pcbStream->Rewind())
	{
		*pfsLog << "...failed returning" << endl;
		return false;
	}
	*pfsLog << "...okay" << endl;

	while (true)
	{
		if (pcbStream->EOBS())
		{
			*pfsLog << "could not find any PTS that suites... return" << endl;
			return false;
		}
		if (pcbStream->Next4Bytes() == ui32StartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
			if (pppPES->IsOkay() && pppPES->PTS())
			{
				if (uiStreamType == VIDEO)
				{
					ui32Next4Bytes=pcbStream->Next4Bytes();
					if ((ui32Next4Bytes == 0x00000100 || ui32Next4Bytes == 0x000001B3) && pppPES->DTS_Available())
					{
						i64PTS=pppPES->PTS();
						*pfsLog << "found PTS " << (double)i64PTS << endl;
						*pfsLog << " waiting... for 0x000001B3" << endl;
					}
					delete pppPES;
				}
				else
				{
					i64PTS=pppPES->PTS();
					*pfsLog << "found PTS " << (double)i64PTS << endl;
					*pfsLog << " audio pts... means doodeldee... gotit!!! -> return" << endl;
					delete pppPES;
					return true;
				}
			}
			else
			{
				delete pppPES;
			}
		}
		if (i64PTS && pcbStream->Next4Bytes() == 0x000001B3 )
		{
			*pfsLog << "got 0x000001B3 after PTS " << (double)i64PTS << "... return" << endl;
			return true;
		}
		else
		{
			pcbStream->MoveByte();
		}
	}
	// you may never be here...
	return false;
}

void CRawFile::RipPrivateStream(CBits* pcbStream, 
							 COBits* pcoOutput, 
							 unsigned int uiStreamID, 
							 __int64 i64StartPTS , 
							 __int64 i64EndPTS)
{
	unsigned __int32	ui32PESStartCode = 0x100 + uiStreamID;
	CPES_Packet*		pppPES=NULL;
	unsigned __int64	pos=0;
	unsigned __int64	LastPTS=0;

	*pfsLog << "CRawFile::RipPrivateStream" << endl;
	cout << "CRawFile::RipPrivateStream" << endl;

	*pfsLog << "Try to rewind...";
	cout << "Try to rewind...";
	// first move to StartPTS
	if (!pcbStream->Rewind())
	{
		*pfsLog << " failed!" << endl;
		cout << " failed!" << endl;
		return;
	}
	*pfsLog	<< " passed!" << endl;
	cout << " passed!" << endl;

	while(!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
			if (pppPES->IsOkay() && pppPES->PTS() >= i64StartPTS)
			{
				LastPTS=pppPES->PTS();
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
					// found this and next
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3,pcbStream,pcoOutput);
					*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
					cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
					delete pppPES;
					break;
				}
				else
				{
					*pfsLog << "this is not a PES_PACKET" << endl;
					delete pppPES;
					pcbStream->MoveByte();
					pcbStream->NextStartCode(MBYTE*10);
				}
			}
			else
			{
				*pfsLog << "PES ist not okay OR PTS is less then StartPTS" << endl;
				delete pppPES;
				pcbStream->MoveByte();
				pcbStream->NextStartCode(MBYTE*10);
			}
		}
		else
		{
			*pfsLog << "this is not the expected startcode!" << endl;
			pcbStream->MoveByte();
			pcbStream->NextStartCode(MBYTE*10);
		}
	}
	
	bool bBitStreamError=false;
	cout << "patience... running..." << endl;
	// now we are here with valid data to process
	while (!pcbStream->EOBS())
	{
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
		if (pppPES->IsOkay())
		{
			if (pppPES->PTS() >= i64EndPTS)
			{
				delete pppPES;
				break;
			}
			else
			{
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
					if (pppPES->PTS() + 180000 < LastPTS)
					{
						i64EndPTS = pppPES->PTS() + (i64EndPTS - i64StartPTS - (LastPTS - i64StartPTS));
						*pfsLog << "PTS overflow..." << endl;
						*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						cout << endl << "PTS overflow..." << endl;
						cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						//it's not possible to know how far we're gone without timing information
						bBitStreamError=false;
					}
					// recovery!!!
					if (bBitStreamError)
					{
						*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						AC3Recovery(pppPES->PTS()-LastPTS,pcoOutput);
						bBitStreamError=false;
					}
					pcbStream->SeekTo(pos);
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3,pcbStream,pcoOutput);
					LastPTS=pppPES->PTS();
				}
				else
				// if not PES then skip these bytes
				{
					bBitStreamError=true;
					*pfsLog << "error recovery" << endl;
					cout << "error recovery" << endl;
					pcbStream->SeekTo(pos+8);
					pcbStream->NextStartCode();
				}
			}
		}
		else
			// if not PES then skip these bytes
		{
			bBitStreamError=true;
			*pfsLog << "error recovery" << endl;
			cout << "error recovery" << endl;
			pcbStream->MoveByte();
			pcbStream->NextStartCode();
		}
		delete pppPES;
	}

	if(pcbStream->EOBS())
	{
		*pfsLog << "EOBS reached!" << endl;
		cout << "EOBS reached!" << endl;
	}
}

void CRawFile::RipAudioStream(CBits* pcbStream, 
							 COBits* pcoOutput, 
							 unsigned int uiStreamID, 
							 __int64 i64StartPTS , 
							 __int64 i64EndPTS)
{
	unsigned __int32	ui32PESStartCode = 0x100 + uiStreamID;
	CPES_Packet*		pppPES=NULL;
	unsigned __int64	pos=0;
	unsigned __int64	LastPTS=0;

	*pfsLog << "CRawFile::RipAudioStream" << endl;
	cout << "CRawFile::RipAudioStream" << endl;

	*pfsLog << "Try to rewind...";
	cout << "Try to rewind...";
	// first move to StartPTS
	if (!pcbStream->Rewind())
	{
		*pfsLog << " failed!" << endl;
		cout << " failed!" << endl;
		return;
	}
	*pfsLog	<< " passed!" << endl;
	cout << " passed!" << endl;

	while(!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
			if (pppPES->IsOkay() && pppPES->PTS() >= i64StartPTS)
			{
				LastPTS=pppPES->PTS();
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
					// found this and next
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3,pcbStream,pcoOutput);
					*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
					cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
					delete pppPES;
					break;
				}
				else
				{
					*pfsLog << "this is not a PES_PACKET" << endl;
					delete pppPES;
					pcbStream->MoveByte();
					pcbStream->NextStartCode(MBYTE*10);
				}
			}
			else
			{
				*pfsLog << "PES ist not okay OR PTS is less then StartPTS" << endl;
				delete pppPES;
				pcbStream->MoveByte();
				pcbStream->NextStartCode(MBYTE*10);
			}
		}
		else
		{
			*pfsLog << resetiosflags(ios::dec) << setiosflags(ios::hex) << (long) pcbStream->Next4Bytes() << " is not the expected startcode!" << endl;
			pcbStream->MoveByte();
			pcbStream->NextStartCode(MBYTE*10);
		}
	}
	
	bool bBitStreamError=false;
	cout << "patience... running..." << endl;
	// now we are here with valid data to process
	while (!pcbStream->EOBS())
	{
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
		if (pppPES->IsOkay())
		{
			if (pppPES->PTS() >= i64EndPTS)
			{
				delete pppPES;
				break;
			}
			else
			{
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
					if (pppPES->PTS() + 180000 < LastPTS)
					{
						i64EndPTS = pppPES->PTS() + (i64EndPTS - i64StartPTS - (LastPTS - i64StartPTS));
						*pfsLog << "PTS overflow..." << endl;
						*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						cout << endl << "PTS overflow..." << endl;
						cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						//it's not possible to know how far we're gone without timing information
						bBitStreamError=false;
					}
					// recovery!!!
					if (bBitStreamError)
					{	
						*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						MPEG_Audio_Recovery(pppPES->PTS()-LastPTS,pcoOutput);
						bBitStreamError=false;
					}
					pcbStream->SeekTo(pos);
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3,pcbStream,pcoOutput);
					LastPTS=pppPES->PTS();
				}
				else
				// if not PES then skip these bytes
				{
					bBitStreamError=true;
					*pfsLog << "error recovery" << endl;
					cout << "error recovery" << endl;
					pcbStream->SeekTo(pos+8);
					pcbStream->NextStartCode();
				}
			}
		}
		else
			// if not PES then skip these bytes
		{
			bBitStreamError=true;
			*pfsLog << "error recovery" << endl;
			cout << "error recovery" << endl;
			pcbStream->MoveByte();
			pcbStream->NextStartCode();
		}
		delete pppPES;
	}

	if(pcbStream->EOBS())
	{
		*pfsLog << "EOBS reached!" << endl;
		cout << "EOBS reached!" << endl;
	}
}

void CRawFile::CopyBytes(unsigned int BytesToCopy, CBits* pcbStream, COBits* pcoOutput)
{
	unsigned int iByte;
	m_RecoveryBuffer.erase(m_RecoveryBuffer.begin(),m_RecoveryBuffer.end());
	while (!pcbStream->EOBS() && BytesToCopy--)
	{
		pcbStream->GetByte(iByte);
		pcoOutput->PutByte(iByte);
		m_RecoveryBuffer.push_back(iByte);
	}
}

void CRawFile::AC3Recovery(long PTSDiff, COBits* pcoOutput)
{
//	for (int i=0; i < m_RecoveryBuffer.size(); i++)
	double iFramesPerS=448000.0 / 8 / 1792; //31,25
	int iFrames = (double)PTSDiff / 90000 * iFramesPerS + 0.5;

	int iBytes=iFrames*1792;
	int i=0;
	while (iBytes--)
	{
		pcoOutput->PutByte(m_RecoveryBuffer[i++]);
		if (i>=m_RecoveryBuffer.size())
			i=0;
	}
}

void CRawFile::MPEG_Audio_Recovery(long PTSDiff, COBits* pcoOutput)
{
//	for (int i=0; i < m_RecoveryBuffer.size(); i++)
	double iFramesPerS=192000.0 / 8 / 576; //31,25
	int iFrames = (double)PTSDiff / 90000 * iFramesPerS + 0.5;

	int iBytes=iFrames*576;
	int i=0;
	while (iBytes--)
	{
		pcoOutput->PutByte(m_RecoveryBuffer[i++]);
		if (i>=m_RecoveryBuffer.size())
			i=0;
	}
}


void CRawFile::RipVideoStream(CBits* pcbStream, 
						   COBits* pcoOutput, 
						   unsigned int uiStreamID, 
						   __int64 i64StartPTS , 
						   __int64 i64EndPTS)
{
	unsigned __int32	ui32PESStartCode = 0x100 + uiStreamID;
	CPES_Packet*		pppPES=NULL;
	unsigned int		uiByte=0;
	bool				PTS_Reached=false;
	STREAM_TYPE			stStreamType=GetStreamType(uiStreamID);
	__int64				i64LastPTS=0, i64ThisPTS=0;
	int nOutput (0);


	// better be good to me
	i64EndPTS--;

	*pfsLog << endl << "CRawFile::RipVideoStream" << endl;
	cout << endl << "CRawFile::RipVideoStream" << endl;
	cout << setiosflags(ios::fixed) ;

	*pfsLog << "rewinding stream";
	cout << "rewinding stream";
	if (pcbStream->Rewind())
	{
		*pfsLog << "...okay";
		cout << "...okay";
	}
	else
	{
		*pfsLog << "...failed!!!";
		cout << "...failed!!!";
	}
	*pfsLog << endl;
	cout << endl;

	*pfsLog << resetiosflags(ios::dec) << setiosflags(ios::hex);
	*pfsLog << "seeking stream start code " << (unsigned int)ui32PESStartCode << endl; 
	cout << resetiosflags(ios::dec) << setiosflags(ios::hex);
	cout << "seeking stream start code " << (unsigned int)ui32PESStartCode << endl;

	while (!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			if (!PTS_Reached)
			{
				pppPES = new CPES_Packet(pcbStream, uiStreamID);
				if (pppPES->IsOkay() && pppPES->PTS() >= i64StartPTS)
				{
					*pfsLog << "reached pts " << (double)pppPES->PTS() << endl;
					PTS_Reached=true;
					*pfsLog << "shit.. i'm video and i have to wait until 0x000001B3...." << endl;
				}
				delete pppPES;
			}
		}
		if (PTS_Reached && pcbStream->NextBits(32)==0x000001B3)
		{
			*pfsLog << "yeah.. i'm video and i reached 0x000001B3...." << endl;
			break;
		}
		pcbStream->MoveByte();
		pcbStream->NextStartCode();
	}
	
	i64LastPTS=i64StartPTS;

	if (pcbStream->EOBS())
	{
		*pfsLog << "EOF reached! No PTS for Code " << (unsigned int)ui32PESStartCode << endl;
		cout << "EOF reached! No PTS for Code " << (unsigned int)ui32PESStartCode << endl;
	}

	*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec);
	cout << resetiosflags(ios::hex) << setiosflags(ios::dec);

	cout << "now ripping until eobs..." << endl;
	*pfsLog << "now ripping until eobs..." << endl;
	cout << setiosflags(ios::fixed);

	while (!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
			if (pppPES->IsOkay())
			{
				i64ThisPTS=pppPES->PTS();
				if (i64ThisPTS + 90000 < i64LastPTS  && i64ThisPTS)
				{
					i64EndPTS = i64ThisPTS + (i64EndPTS - i64StartPTS - (i64LastPTS - i64StartPTS));
					*pfsLog << "PTS overflow..." << endl;
					*pfsLog << "Max PTS=" << double(i64EndPTS) << endl;
					cout << endl << "PTS overflow..." << endl;
					cout << "Max PTS=" << double(i64EndPTS) << endl;
				}
				if (!(nOutput++ % 100))
					cout << " at PTS=" << (double) ((i64ThisPTS / 10000) * 10000) << (char)13;
				if (i64ThisPTS > i64EndPTS)
				{
					cout << " at PTS=" << (double)i64ThisPTS << (char)13 << endl;
					*pfsLog << " at PTS=" << (double)i64ThisPTS << endl;
					*pfsLog << "Stream ripped successfully..." << endl;
					cout << "Stream ripped successfully..." << endl;
					delete pppPES;
					break;
				}
				i64LastPTS=i64ThisPTS;
			}
			// not okay...
			else
			{
				pcbStream->GetByte(uiByte);
				pcoOutput->PutByte(uiByte);
			}
			delete pppPES;
		}
		else
		{
			pcoOutput->PutByte(pcbStream->GetByte());
		}
	}
	if (pcbStream->EOBS())
	{
		*pfsLog << "EOBS Reached..." << endl;
		cout << "EOBS Reached..." << endl;
	}
	pcoOutput->Flush();
	*pfsLog << "return" << endl;
}


void CRawFile::RipStreams(unsigned int iStartSeconds,unsigned int iSeconds)
{
	CBits*					cbVideoStream=NULL;
	CBits*					cbAudioStream=NULL;
	unsigned int			uiVideoStreamID=0;
	unsigned int			uiAudioStreamID=0;

	COBits*			coOut=NULL;
	__int64			i64PTSAudio=0, i64PTSVideo=0, i64FirstPTS=0, i64LastPTS=0;
	int				i=0;

	DWORD			tick=0,tack=0;

	bool			bResult=false;

	Dump(&cout,false);
	Dump(pfsLog,false);

	cout.precision(0);
	cout << "analizing streams..." << endl;


	cbVideoStream=new CBits(m_sDrivename,m_sfdFile.iVideoFatEntry);
	GetStreamIDs(*cbVideoStream, uiVideoStreamID);
	cbVideoStream->Rewind();
	
	if (m_bRipAudio)
	{
		cbAudioStream=new CBits(m_sDrivename,m_sfdFile.iAudioFatEntry);
		GetStreamIDs(*cbAudioStream, uiAudioStreamID);
		cbAudioStream->Rewind();
	}

	cout << setiosflags(ios::hex) ;
	*pfsLog << setiosflags(ios::hex) ;

	*pfsLog << "Videostream=";
	cout << "Videostream=";

	cout << uiVideoStreamID << " ";
	*pfsLog << uiVideoStreamID << " ";

	cout << endl;
	*pfsLog << endl;
	if (m_bRipAudio)
	{
		cout << "Audiostream=";
		*pfsLog << "Audiostream=";
		cout << uiAudioStreamID << " ";
		*pfsLog << uiAudioStreamID << " ";
		cout << endl;
		*pfsLog << endl;
	}

	if ((uiAudioStreamID || !m_bRipAudio) && uiVideoStreamID)
	{
		if (m_bRipAudio)
		{
			cout << "try to find first pts for Audio..." << endl;
			GetFirstPTS(cbAudioStream, uiAudioStreamID, i64PTSAudio);
			cout << setiosflags(ios::fixed) ;
			cout << "Audiostream PTS=" << (double)i64PTSAudio << endl;
		}
		else
		{
			i64PTSAudio = 0xFFFFFFFFFFFFFFF;
		}

		cout << "try to find first pts for Video..." << endl;
		GetFirstPTS(cbVideoStream, uiVideoStreamID, i64PTSVideo);
		cout << setiosflags(ios::fixed) ;
		cout << "Videostream PTS=" << (double)i64PTSVideo << endl;

		i64FirstPTS = min(i64PTSAudio, i64PTSVideo) + (unsigned __int64)iStartSeconds * 90000;
		i64LastPTS = i64FirstPTS + 90000 * iSeconds;

		i64FirstPTS = i64PTSVideo + (unsigned __int64)iStartSeconds * 90000;

		cout << "First PTS=" << double(i64FirstPTS) << endl;
		cout << "Max   PTS=" << double(i64LastPTS) << endl;
		cout << endl;

		*pfsLog << "First PTS=" << double(i64FirstPTS) << endl;
		*pfsLog << "Max   PTS=" << double(i64LastPTS) << endl;
		*pfsLog << endl;
		if (m_bRipAudio)
		{
 
			cout << "now ripping audio" << endl;
			tick = GetTickCount();
			if(GetStreamType(uiAudioStreamID)==AUDIO)
			{
				coOut=new COBits(Filename().append(".mp2"), bResult);
				RipAudioStream(cbAudioStream,coOut,uiAudioStreamID,i64FirstPTS,i64LastPTS);
			}
			else
			{
				coOut=new COBits(Filename().append(".ac3"), bResult);
				RipPrivateStream(cbAudioStream,coOut,uiAudioStreamID,i64FirstPTS,i64LastPTS);
			}
			tack = GetTickCount()-tick;
			cout << "... it took " << dec << (int)tack / 1000 << " seconds..." << endl;
			*pfsLog << "... it took " << dec << (int)tack / 1000 << " seconds..." << endl;
			delete coOut;
		}

		cout << "now ripping video" << endl;
		tick = GetTickCount();
		coOut=new COBits(Filename().append(".m2v"),bResult);
		RipVideoStream(cbVideoStream,coOut,uiVideoStreamID,i64FirstPTS,i64LastPTS);
		// Sequence END Code
		coOut->PutByte(0x00);
		coOut->PutByte(0x00);
		coOut->PutByte(0x01);
		coOut->PutByte(0xb7);
		coOut->Flush();
		delete coOut;

		tack = GetTickCount()-tick;
		cout << setiosflags(ios::dec) ;
		cout << "... it took " << dec << (int)tack / 1000 << " seconds..." << endl;
		*pfsLog << "... it took " << dec << (int)tack / 1000 << " seconds..." << endl;
	}
	else
	{
		cout << "can't go on..." << endl;
		*pfsLog << "can't go on..." << endl;
	}
	cout << "cleaning up" << endl;
	delete cbVideoStream;
	if (m_bRipAudio)
	{
		delete cbAudioStream;
	}
}

void CRawFile::Dump(ostream *posOut, bool bWait)
{
	*posOut << "RecordNo    =" << m_sfdFile.iRecordNo << endl;
	*posOut << "Station     =" << m_sfdFile.cStationName << endl;
	*posOut << "Clip        =" << m_sfdFile.cClipName << endl;
	*posOut << "Date        =" 
		<< m_sfdFile.iRecordDay 
		<< "."
		<< m_sfdFile.iRecordMonth 
		<< "."
		<< m_sfdFile.iRecordYear
		<< " "
		<< m_sfdFile.iRecordHour
		<< ":"
		<< m_sfdFile.iRecordMinute
		<< ":"
		<< m_sfdFile.iRecordSecond
		<< endl;
	*posOut << "Length      =" 
		<< m_sfdFile.iLengthHours 
		<< ":"
		<< m_sfdFile.iLengthMinutes
		<< ":"
		<< m_sfdFile.iLengthSeconds
		<< endl;
	*posOut << "Length in s =" 
		<< m_sfdFile.iLengthHours * 3600 + m_sfdFile.iLengthMinutes * 60 + m_sfdFile.iLengthSeconds 
		<< endl;
	*posOut << "SkipIndex   =" << endl;
	int i=0;
	while(i<MAX_SKIP_INDEX)
	{
//		if (m_sfdFile.ssiSkipIndex[i].iStart==-1 || m_sfdFile.ssiSkipIndex[i].iStop==-1)
//			break;
		if (m_sfdFile.ssiSkipIndex[i].iStart!=-1 && m_sfdFile.ssiSkipIndex[i].iStop!=-1)
		{
		cout 
			<< "start=" 
			<< m_sfdFile.ssiSkipIndex[i].iStart
			<< " stop=" 
			<< m_sfdFile.ssiSkipIndex[i].iStop
			<< endl;
		}
		i++;
	}
	*posOut << "VClusters   =" 
		<< m_sfdFile.iVideoClusterCount
		<< " means max " 
		<< m_sfdFile.iVideoClusterCount * 100 
		<< " MByte"
		<< endl;
	*posOut << "AClusters   =" 
		<< m_sfdFile.iAudioClusterCount
		<< " means max " 
		<< m_sfdFile.iAudioClusterCount * 100 
		<< " MByte"
		<< endl;
	*posOut << "VClusters   =" ;
	g_pfatFAT->DumpClusters(posOut, m_sfdFile.iVideoFatEntry);
	*posOut << "AClusters   =" ;
	g_pfatFAT->DumpClusters(posOut, m_sfdFile.iAudioFatEntry);
	*posOut << endl;

	if (bWait) 
		getchar();
}

void CRawFile::GetClipName(CString &s)
{
	s = m_sfdFile.cClipName;
}

void CRawFile::GetStationName(CString &s)
{
	s = m_sfdFile.cStationName;
}
