#include "stdafx.h"
#include "ibits.h"
#include <iostream>
#include <iomanip>
using namespace std;

CIBits::CIBits(const string &Drivename, unsigned long FatEntry)
{
	m_hFile=INVALID_HANDLE_VALUE;
	m_bBuffer=NULL;
	m_dwBufferCount=0;
	m_lClusterIndex=0;
	m_lBufferIndex=0;
	m_lByteIndex=0;
	m_lBitIndex=0;
	m_bEOBS=false;

	if (OpenDrive(m_hFile, Drivename))
	{
		g_pfatFAT->GetClusters(FatEntry,m_lvecClusters);
		m_bBuffer=new BYTE [BYTES_PER_BUFFER];
		memset (m_bBuffer, 0, BYTES_PER_BUFFER);
		m_dwBufferCount=0;
		Rewind();
	}
};

CIBits::~CIBits()
{
	if (m_hFile != INVALID_HANDLE_VALUE)
		CloseHandle(m_hFile);
	delete []m_bBuffer;
}


void CIBits::BitPointer(unsigned __int64 &ThisValue)
{
	ThisValue = ((__int64)m_lClusterIndex * BYTES_PER_CLUSTER 
						+ (m_lBufferIndex-1) * BYTES_PER_BUFFER 
						+ m_lByteIndex) * 8 
						+ 8 - m_lBitIndex;
}
unsigned __int64 CIBits::BitPointer()
{
	return ((__int64)m_lClusterIndex * BYTES_PER_CLUSTER 
						+ (m_lBufferIndex-1) * BYTES_PER_BUFFER 
						+ m_lByteIndex) * 8 
						+ 8 - m_lBitIndex;
}
       
bool CIBits::ByteAligned()
{
	return (m_lBitIndex==8);
}

bool CIBits::EOBS()
{
	return m_bEOBS;
}
/*read Number of bits from the bit stream */
void CIBits::ClrEOBS()
{
	m_bEOBS = false;
}

unsigned int CIBits::Get1Bit()
{
	unsigned int uiBit;
	uiBit = (m_bBuffer[m_lByteIndex] & BitMasks[m_lBitIndex - 1]) >> (m_lBitIndex - 1);
	m_lBitIndex--;
	if (!m_lBitIndex)
	{
		IncByteIndex();
		m_lBitIndex=8;
	}
	return uiBit;
}

void CIBits::Get1Bit(unsigned int &Result)
{
	if (m_bEOBS)
		return;

	Result = (m_bBuffer[m_lByteIndex] & BitMasks[m_lBitIndex - 1]) >> (m_lBitIndex - 1);
	m_lBitIndex--;
	if (!m_lBitIndex)
	{
		IncByteIndex();
		m_lBitIndex=8;
	}
}

unsigned __int32 CIBits::GetBits(unsigned int Number)
{
	unsigned int uiValue = 0;
	int i = Number;
	unsigned int j;

	// Optimize: we are on byte boundary and want to read multiple of bytes!
	if ((m_lBitIndex == 8) && ((Number & 7) == 0))
	{
		i = Number >> 3;
		while (i > 0 && !m_bEOBS)
		{
			uiValue = (uiValue << 8) | m_bBuffer[m_lByteIndex];
			IncByteIndex();
			i--;
		}
		return uiValue;
	}

	while (i-- > 0)
	{
		j = Get1Bit();
		uiValue = (uiValue << 1) | j;
	}
	return uiValue;
}


void CIBits::GetByte(unsigned int &Byte)
{
	if (m_bEOBS)
	{
		Byte=0;
	}
	else
	{
		Byte=m_bBuffer[m_lByteIndex];
		IncByteIndex();
	}
	return;
}

unsigned int CIBits::GetByte()
{
	if (m_bEOBS)
	{
		return 0;
	}
	else
	{
		unsigned int uiRetValue=m_bBuffer[m_lByteIndex];
		IncByteIndex();
		return uiRetValue;
	}
}

inline void CIBits::IncByteIndex()
{
	if (++m_lByteIndex == (long)m_dwBufferCount)
	{
		if (m_dwBufferCount == BYTES_PER_BUFFER)
		{
			if (!RefillBuffer())
			{
				m_bEOBS=true;
			}
		}
		else
		{
			m_bEOBS = true;
		}
	}
}


void CIBits::MoveByte()
{
	IncByteIndex();
}

void CIBits::MoveBytes(unsigned int NumberOfBytes)
{
	while (NumberOfBytes-- > 0)
		IncByteIndex();
}

unsigned __int32 CIBits::NextBits(unsigned int NumberOfBits)
{
	unsigned __int32	uiValue;
	unsigned __int64	i64OldBitPointer;
	BitPointer(i64OldBitPointer);
	uiValue=GetBits(NumberOfBits);
	SeekTo(i64OldBitPointer);
	return uiValue;
}

unsigned __int32 CIBits::Next4Bytes()
{
	unsigned __int32	uiValue;
	unsigned __int64	i64OldBitPointer;
	BitPointer(i64OldBitPointer);
	uiValue=GetByte() << 24 |  GetByte() << 16 | GetByte() << 8 | GetByte();
	SeekTo(i64OldBitPointer);
	return uiValue;
}

bool CIBits::RefillBuffer()
{
	BOOL			bReturn;
	__int64			liSector;
  CString s;

	if (m_lBufferIndex * SECTORS_PER_BUFFER == SECTORS_PER_CLUSTER)
	{
		m_lClusterIndex++;
		if (m_lClusterIndex >= (long) m_lvecClusters.size())
		{
      s.Format ("RefillBuffer() - m_lClusterIndex = %d >= (long) m_lvecClusters.size() = %d failed\n",m_lClusterIndex, (long) m_lvecClusters.size());
      Output (s);
			m_bEOBS=true;
			return false;
		}
		else
		{
			*pfsLog << "goto next cluster " << m_lClusterIndex << "/" << m_lvecClusters[m_lClusterIndex] << endl;
			liSector=(__int64)(m_lvecClusters[m_lClusterIndex]-1) * SECTORS_PER_CLUSTER + SECTOR_OFFSET;
			m_lBufferIndex=0;
			if (!SetToPhysicalSector(m_hFile, liSector))
			{
				m_bEOBS=true;
				return false;
			}
			else
			{
				return RefillBuffer();
			}
		}
	}

	bReturn=ReadFile(m_hFile,m_bBuffer,BYTES_PER_BUFFER,&m_dwBufferCount,NULL);
	if (!bReturn || m_dwBufferCount==0)
	{
		m_bEOBS=true;
		return false;
	}
	m_lBitIndex=8;
	m_lByteIndex=0;
	m_lBufferIndex++;
	return true;
};

bool CIBits::Rewind()
{
	
	__int64		liSector;
  CString s;
  Output ("CIBits::Rewind()\n");
	liSector=(__int64)(m_lvecClusters[0]-1) * SECTORS_PER_CLUSTER + SECTOR_OFFSET;
	if (!SetToPhysicalSector(m_hFile, liSector))
	{
		m_bEOBS=true;
		return false;
	}
	m_lClusterIndex=0;
	m_lBufferIndex=0;
	m_lByteIndex=0;
	m_lBitIndex=8;
	if (!RefillBuffer())
	{
    s.Format ("RefillBuffer() failed\n");
    Output (s);
		m_bEOBS=true;
		return false;
	}
	m_bEOBS=false;
  Output ("end of CIBits::Rewind()\n");
	return true;
}


bool CIBits::SeekTo(unsigned __int64 ToBitPosition)
{
	long		lClusters;
	long		lBuffersInCluster;

	if (m_bEOBS)
		return false;
	lClusters = (long) (ToBitPosition / BITS_PER_CLUSTER);
	if (lClusters >= (long) m_lvecClusters.size())
	{
		m_bEOBS=true;
		return false;
	}

	ToBitPosition -= (__int64) lClusters * BITS_PER_CLUSTER;
	lBuffersInCluster = (long) ToBitPosition / BITS_PER_BUFFER;

	ToBitPosition -= (__int64) lBuffersInCluster * BITS_PER_BUFFER;

	if (lBuffersInCluster != (m_lBufferIndex-1)  || lClusters != m_lClusterIndex)
	{
		if (!SetToPhysicalSector(m_hFile, 
			(__int64)(m_lvecClusters[lClusters] - 1) * SECTORS_PER_CLUSTER 
			+ lBuffersInCluster * SECTORS_PER_BUFFER 
			+ SECTOR_OFFSET))
		{
			m_bEOBS=true;
			return false;
		}
		m_lClusterIndex = lClusters;
		m_lBufferIndex = lBuffersInCluster;
		if (!RefillBuffer())
		{
			m_bEOBS=true;
			return false;
		}
	}	

	m_lByteIndex = (long)(ToBitPosition >> 3);
	ToBitPosition -= (m_lByteIndex << 3);
	m_lBitIndex = 8 - (long)ToBitPosition;
	return true;
}

int CIBits::SetPositionRaw(__int64 i64ToPosition)
{
	return SetToPhysicalSector(m_hFile, i64ToPosition); 
 
}

void CIBits::Output (const char * s)
{
	bool bNewline (false);
	char buf[255];
	strcpy (buf, s);
	if (buf[strlen(buf)-1] == '\n') 
	{
		buf[strlen(buf) - 1] = 0;
		bNewline = true;
	}
	*pfsLog << buf;
	if (bNewline) *pfsLog << endl;
}


