#ifndef PVR_MPEG_H
#define PVR_MPEG_H
#pragma once

#include "general.h"
#include "ibits.h"

// MPEG stuff

#define	START_CODE_PREFIX				0x000001

// PES packets stream IDs and assignment
#define PROGRAM_STREAM_MAP				0xBC
#define PRIVATE_STREAM_1				0xBD
#define PADDING_STREAM					0xBE
#define PRIVATE_STREAM_2				0xBF
#define ISO_IEC_AUDIO_STREAM			0xC0
#define ISO_IEC_VIDEO_STREAM			0xE0
#define ECM_STREAM						0xF0
#define EMM_STREAM						0xF1
#define DSMCC_STREAM					0xF2
#define	ISO_IEC_13522_STREAM			0xF3
#define ITU_T_A_STREAM					0xF4
#define ITU_T_B_STREAM					0xF5
#define ITU_T_C_STREAM					0xF6
#define ITU_T_D_STREAM					0xF7
#define ITU_T_E_STREAM					0xF8
#define ANCILLARY_STREAM				0xF9
#define PROGRAM_STREAM_DIRECTORY		0xFF

#define AC3_FRAME_START					0x0B77
#define AC3_FRAME_END					0x0BFF
#define MAX_PTS_33BIT					0x200000000
#define B_FRAME               3
enum STREAM_TYPE
{
	OTHER,
	PRIVATE_1,
	PRIVATE_2,
	AUDIO,
	VIDEO
};

extern const STREAM_TYPE STREAM_TYPES[0x100];


#define VIDEO_STREAMS					0x10
#define AUDIO_STREAMS					0x20

enum TRICK_MODE_CONTROL
{
	FAST_FORWARD,
	SLOW_MOTION,
	FREEZE_FRAME,
	FAST_REVERSE,
	SLOW_REVERSE
	// remaining are reserved
};

//------------------------------------------------------------------
bool NextStartCode(CIBits &InStream);
bool NextStartCode(CIBits &InStream, unsigned long MaxBytesToSearch);



extern inline STREAM_TYPE GetStreamType(unsigned int StreamID);
__int64 IsPES_Packet(CIBits &InStream, unsigned int StreamID, __int32 * pnLength = NULL);

// PES and PACK

struct sPack_Header
{
private:
	unsigned int	SCRB_32_30;
	unsigned int	SCRB_29_15;
	unsigned int	SCRB_14_00;
	unsigned int	Pack_Stuffing_Length;

public:
	bool PackHeader(CIBits* pStream)
	{
		if (pStream->GetBits(32) != 0x000001BA)
			return false;
		if (pStream->GetBits(2) != 0x1)
			return false;

		SCRB_32_30=pStream->GetBits(3);
		if (!pStream->Get1Bit())
			return false;

		SCRB_29_15=pStream->GetBits(15);
		if (!pStream->Get1Bit())
			return false;

		SCRB_14_00=pStream->GetBits(15);
		if (!pStream->Get1Bit())
			return false;

		pStream->GetBits(9);
		if (!pStream->Get1Bit())
			return false;


		pStream->GetBits(22);
		if (!pStream->Get1Bit())
			return false;
		if (!pStream->Get1Bit())
			return false;

		pStream->GetBits(3);

		Pack_Stuffing_Length=pStream->GetBits(3);
		for (unsigned int i=0;i<Pack_Stuffing_Length;i++)
		{
			if (pStream->GetByte() != 0xFF)
				return false;
		}
/*
		if (pStream->Next4Bytes() == 0x000001BB)
		{
			// raus!!!
			int a=0,b=0;
			b=b/a;
		}
*/
		return true;
	};
	__int64 SCRB()
	{
		return ((__int64)SCRB_32_30 << 30) + (SCRB_29_15 << 15) + SCRB_14_00;
	}
};


class CPES_Packet
{
public:
	CPES_Packet(CIBits* pbcBitStream, unsigned int uiStreamID);

	__int64			PTS();
	__int64			DTS();
	bool			DTS_Available();
	bool			IsOkay();
	unsigned int	PES_Packet_Length();
	unsigned int	PES_Header_Data_Length();
private:
	bool			GetPES_Packet();

	CIBits*			m_pcbBitStream;
	unsigned int	m_uiStreamID;
	bool			m_bIsOkay;
	__int64			m_PTS;
	__int64			m_DTS;
	sPack_Header	m_Pack_Header;


	unsigned int	Stream_ID;
	unsigned int	m_PES_Packet_Length;
	unsigned int	Data_Alignment_Indicator;
	unsigned int	PTS_DTS_Flags;
	unsigned int	ESCR_Flag;
	unsigned int	ES_Rate_Flag;
	unsigned int	DSM_Trick_Mode_Flag;
	unsigned int	Additional_Copy_Info_Flag;
	unsigned int	PES_CRC_Flag;
	unsigned int	PES_Extension_Flag;
	unsigned int	m_PES_Header_Data_Length;
	unsigned int	PES_Private_Data_Flag;
	unsigned int	Pack_Header_Field_Tag;
	unsigned int	Program_Packet_Sequence_Counter_Flag;
	unsigned int	P_STD_Buffer_Flag;
	unsigned int	PES_Extension_Flag_2;
	unsigned int	Pack_Field_Length;
	unsigned int	PES_Extension_Field_Length;
	unsigned int	Trick_Mode_Control;
	unsigned int	PTS_32_30;
	unsigned int	PTS_29_15;
	unsigned int	PTS_14_00;
	unsigned int	DTS_32_30;
	unsigned int	DTS_29_15;
	unsigned int	DTS_14_00;

	unsigned int	m_Real_PES_Header_Data_Length;

};



#endif