#include "stdafx.h"
#include "general.h"
#include "obits.h"
#include <cmath>
using namespace std;

COBits::COBits(const string &Filename, bool &Okay)
{

	*pfsLog << "COBits::COBits" << endl;
	*pfsLog << Filename << endl;
  OFSTRUCT of;
  CString sFilename (Filename.c_str()), sExt, s, sNum;
	int i (0);
	m_hFile = INVALID_HANDLE_VALUE;
  sExt = sFilename.Mid (sFilename.ReverseFind ('.'));
  sFilename = sFilename.Left (sFilename.ReverseFind ('.'));
  s = sFilename + sExt;
  while (OpenFile (s, &of, OF_EXIST) != HFILE_ERROR && sFilename != "PRV")
  {
    sNum.Format ("_%03d", i++);
    s = sFilename + sNum + sExt;
  }
  if (sFilename == "PRV") s = "PRV.m2v";
  m_sFilename = s;
	m_hFile=CreateFile(s,
		GENERIC_WRITE,
		FILE_SHARE_READ,
		NULL,
		CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,
		NULL);

	if (m_hFile != INVALID_HANDLE_VALUE)
	{
		*pfsLog << "filehandle okay" << endl;
	}
	else
	{
		*pfsLog << "filehandle NOT okay!!!" << endl;
		Okay=false;
		return;
	}

	m_iByteIndex=0;
	m_iBitIndex=8;
	m_bBuffer=NULL;
	m_bBuffer=new unsigned char [OBITS_BUFFER_SIZE];
	if (!m_bBuffer)
	{
		*pfsLog << "could net get memory for buffer" << endl;
		Okay=false;
		return;
	}
	ClearBuffer();
	for (i=0;i<32;i++)
		m_ui32BitMasks[i]=(unsigned __int32)pow(2,i);
	Okay=true;
}

COBits::~COBits()
{
	if (m_hFile != INVALID_HANDLE_VALUE)
	{
		this->Flush();
		CloseHandle(m_hFile);
	}
	if (m_bBuffer)
		delete m_bBuffer;
}

void COBits::PutByte(unsigned int &Byte)
{
	if (m_iByteIndex > OBITS_BUFFER_SIZE || m_iByteIndex < 0)
		m_iByteIndex = 0;
	if (m_iBitIndex != 8)
	{
		m_iBitIndex=8;
		m_iByteIndex++;
		if (m_iByteIndex == OBITS_BUFFER_SIZE)
			this->Flush();
	}
	m_bBuffer[m_iByteIndex++]=Byte;
	if (m_iByteIndex == OBITS_BUFFER_SIZE)
		this->Flush();
}

void COBits::PutByte(const int Byte)
{
	if (m_iBitIndex != 8)
	{
		m_iBitIndex=8;
		m_iByteIndex++;
		if (m_iByteIndex == OBITS_BUFFER_SIZE)
			this->Flush();
	}
	m_bBuffer[m_iByteIndex++]=Byte;
	if (m_iByteIndex == OBITS_BUFFER_SIZE)
		this->Flush();
}

void COBits::Put1Bit(const unsigned int Bit)
{
	if (Bit)
		m_bBuffer[m_iByteIndex] += BitMasks[m_iBitIndex - 1];
	m_iBitIndex--;
	if (!m_iBitIndex)
	{
		m_iBitIndex=8;
		m_iByteIndex++;
		if (m_iByteIndex == OBITS_BUFFER_SIZE)
			this->Flush();
	}
}

void COBits::PutBits(const unsigned __int32 Value,unsigned int NumberOfBits)
{

	while (NumberOfBits)
	{
		// ByteAligned and Bytes to store
		if (m_iBitIndex==8 && !(NumberOfBits % 8))
		{
			m_bBuffer[m_iByteIndex] = (byte)((Value & ByteMasks[NumberOfBits / 8]) >> (NumberOfBits - 8));
			NumberOfBits -= 8;

			m_iByteIndex++;
			if (m_iByteIndex == OBITS_BUFFER_SIZE)
				this->Flush();
		}
		else
		{
			NumberOfBits--;
			this->Put1Bit((Value & m_ui32BitMasks[NumberOfBits]) >> (NumberOfBits));
		}
	}
}

void COBits::ClearBuffer()
{
	memset(m_bBuffer,0,OBITS_BUFFER_SIZE * sizeof(BYTE));
}
CString & COBits::GetFilename ()
{
  return m_sFilename;
}

void COBits::Flush()
{
	DWORD	dwBytesWritten;

	if (m_iBitIndex != 8)
		m_iByteIndex++;
	if (!m_iByteIndex)
		return;
	if (!WriteFile(m_hFile, m_bBuffer, m_iByteIndex, &dwBytesWritten, NULL))
	{
		*pfsLog << "COBits::Flush WriteFile failed!" << endl;
	}
	m_iByteIndex=0;
	m_iBitIndex=8;
	ClearBuffer();
}

bool COBits::ByteAligned()
{
	return (m_iBitIndex==8);
}