#include "stdafx.h"
#include "pvrfile.h"
#include "mpeg.h"
#include "general.h"
#include <cmath>
#include <iomanip>
#include <iostream>
using namespace std;

CPVRFile::CPVRFile(const string &Drivename, unsigned int FileNo)
: m_nFirmwareType(0)
, m_nLastClusterOffset(0)
, m_i64VideoMaxPos(0)
, m_dAspectRatio(4.0/3.0)
, m_dOffsetTime(0)
, m_dRipTime(0)
, m_dLenInSeconds(0)
{
	HANDLE			hFile;
	DWORD			dwBytesRead;
  CFile fDeb;
  CTime t;
	*pfsLog << "CPVRFile::CPVRFile(char *Drivename, int FileNo)" << endl;
  m_i64PTSAudio = -1;
  m_i64PTSVideo = -1;

	m_nFileNo = FileNo;
	m_sDrivename=Drivename;
  t.GetLocalTm ();
  m_pi64ClusterPositions = NULL;
  if (Drivename == "REC")
  {
    if (fDeb.Open ("driveheader.rec", CFile::modeRead | CFile::typeBinary))
    {
      fDeb.Seek ((TOC_OFFSET_SECTORS + FileNo) * BYTES_PER_SECTOR, CFile::begin);
      fDeb.Read (&m_sfdFile, sizeof (m_sfdFile));
  	  memcpy (&m_sfdFile_old, &m_sfdFile, sizeof (m_sfdFile));
      fDeb.Close ();
    }
    if (!g_bFirmware)
    {
      m_nFirmwareType = ID_189;
      memcpy (&m_sfdFile.iRecordYear, &m_sfdFile_old.iRecordYear, sizeof (m_sfdFile) - 100);
      memset (m_sfdFile.cClipName, 0x20, sizeof (m_sfdFile.cClipName));
      memcpy (m_sfdFile.cClipName, m_sfdFile_old.cClipName, sizeof (m_sfdFile_old.cClipName));
      memset (m_sfdFile.cStationName, 0x20, sizeof (m_sfdFile.cStationName));
      memcpy (m_sfdFile.cStationName, m_sfdFile_old.cStationName, sizeof (m_sfdFile_old.cStationName));
      m_sfdFile.cFWMarker = 0;
    }
    else 
    {
      m_nFirmwareType = ID_190;
    }
  }
  else
  {
	  if (OpenDrive(hFile, Drivename))
	  {
		  SetToPhysicalSector(hFile, TOC_OFFSET_SECTORS + FileNo);
		  ReadFile(hFile, &m_sfdFile,sizeof(m_sfdFile),&dwBytesRead,NULL);
  	  memcpy (&m_sfdFile_old, &m_sfdFile, sizeof (m_sfdFile));
      if (!g_bFirmware)
      {
        m_nFirmwareType = ID_189;
        memcpy (&m_sfdFile.iRecordYear, &m_sfdFile_old.iRecordYear, sizeof (m_sfdFile) - 100);
        memset (m_sfdFile.cClipName, 0x20, sizeof (m_sfdFile.cClipName));
        memcpy (m_sfdFile.cClipName, m_sfdFile_old.cClipName, sizeof (m_sfdFile_old.cClipName));
        memset (m_sfdFile.cStationName, 0x20, sizeof (m_sfdFile.cStationName));
        memcpy (m_sfdFile.cStationName, m_sfdFile_old.cStationName, sizeof (m_sfdFile_old.cStationName));
        m_sfdFile.cFWMarker = 0;
      }
      else 
      {
        m_nFirmwareType = ID_190;
      }
     	CloseHandle(hFile);
    }
   	else
	  {
		  *pfsLog << "could not open drive " << Drivename << " for file " << FileNo << endl;
	  }
	}
  m_dLenInSeconds = m_sfdFile.iLengthHours * 360000 + m_sfdFile.iLengthMinutes * 6000 + m_sfdFile.iLengthSeconds * 100;
	m_RecordNoOkay=(FileNo==m_sfdFile.iRecordNo);
  m_pBuffer = new BYTE[4 * MBYTE];
	
}

CPVRFile::CPVRFile (const string &Drivename, sFileDescription * psfdFile)
: m_nFirmwareType(0)
, m_nLastClusterOffset(0)
, m_i64VideoMaxPos(0)
, m_dAspectRatio(4.0/3.0)
, m_dOffsetTime(0)
, m_dRipTime(0)
, m_dLenInSeconds(0)
{
  m_pi64ClusterPositions = NULL;
  m_dAspectRatio = 4.0/3.0;
	memcpy (&m_sfdFile, psfdFile, sizeof (m_sfdFile));
	memcpy (&m_sfdFile_old, psfdFile, sizeof (m_sfdFile));
  m_i64PTSAudio = -1;
  m_i64PTSVideo = -1;
  if (m_sfdFile.cFWMarker == 0)
  {
      m_nFirmwareType = ID_189;
      memcpy (&m_sfdFile.iRecordYear, &m_sfdFile_old.iRecordYear, sizeof (m_sfdFile) - 100);
      memset (m_sfdFile.cClipName, 0x20, sizeof (m_sfdFile.cClipName));
      memcpy (m_sfdFile.cClipName, m_sfdFile_old.cClipName, sizeof (m_sfdFile_old.cClipName));
      memset (m_sfdFile.cStationName, 0x20, sizeof (m_sfdFile.cStationName));
      memcpy (m_sfdFile.cStationName, m_sfdFile_old.cStationName, sizeof (m_sfdFile_old.cStationName));
      m_sfdFile.cFWMarker = 0;
  }
  else 
  {
    m_nFirmwareType = ID_190;
  }
	m_sDrivename=Drivename;
  m_dLenInSeconds = m_sfdFile.iLengthHours * 360000 + m_sfdFile.iLengthMinutes * 6000 + m_sfdFile.iLengthSeconds * 100;
  m_pBuffer = new BYTE[4 * MBYTE];
}

CPVRFile::~CPVRFile()
{
  delete m_pBuffer;
}

string CPVRFile::Filename()
{
	char	buf[4];
	string	Filename ("");
	ltoa((long)m_sfdFile.iRecordNo,buf,10);
	//Filename.append(buf);
	//Filename=Filename.substr(Filename.size()-3,Filename.size());
	//Filename.append(" - ");
	if (strlen (m_cFilename))
	{
		Filename.append (m_cFilename);
	}
	else
	{
		Filename.append(m_sfdFile.cClipName);
	}
	string s="?\\*\"<>|/:";

	for (unsigned int n=0;n<s.size();n++)
	{
		string::size_type pos=Filename.find(s.substr(n,1),0);
		while (pos<Filename.size())
		{
			Filename.replace(pos,1," ");
			pos=Filename.find(s.substr(n,1),0);
		}
	}
	return Filename;
}


// Sucht nach multiplen Streams in dem Cluster
bool CPVRFile::SearchStreamIDs(CIBits &InStream, unsigned int &StreamID, int &nSkippedBuffers)
{
	int					i, nCount (0);
	unsigned __int64	ui64BitPointer;
	unsigned __int32	ui32Next4Bytes, nStream, nLastStreamID (0);
	nSkippedBuffers = 0;
	StreamID=0;
	char buffer[40];

	__int64 i64CurrentPTS, i64LastPTS;

	*pfsLog << "SearchStreamIDs (split)" << endl;
	*pfsLog << "try to rewind";
	if (!InStream.Rewind())
	{
		*pfsLog << "... failed..." << endl;
		return false;
	}
	*pfsLog << "... okay" << endl;
	Output ("trying to find Stream IDs (clipped file)");
	ui64BitPointer=0;
	if (m_nRipType == RIP_TYPE_AUDIO) m_IDsAudio.m_nNumberFound = 0;
	if (m_nRipType == RIP_TYPE_VIDEO) m_IDsVideo.m_nNumberFound = 0;

	while (NextStartCode(InStream, MBYTE * 10 * 10) && ui64BitPointer < (__int64) MBYTE * 80 * 10)
	{
#ifdef _WINPVR
		if (*m_pbCancel)
		{
			return false;
		}
#endif
		ui32Next4Bytes=InStream.Next4Bytes();
		//ui64BitPointer=InStream.NextBits(32);
		for (i=0xBD; i < 0xF0;i++)
		{
			switch(GetStreamType(i))
			{
			case OTHER:
				break;
			default:
				i64CurrentPTS = IsPES_Packet(InStream,i);

				if (i64CurrentPTS)
				{
					nStream = GetStreamType (i);
					if ((nStream == AUDIO || nStream == PRIVATE_2 || nStream == PRIVATE_1)
						&& m_nRipType == RIP_TYPE_AUDIO && (i != nLastStreamID 
						|| i64abs (i64CurrentPTS, i64LastPTS) > 0x80000))
					{
						m_IDsAudio.m_nIDs[m_IDsAudio.m_nNumberFound] = i;
						m_IDsAudio.m_nBuffersSkipped[m_IDsAudio.m_nNumberFound] = nSkippedBuffers;
						m_IDsAudio.m_i64CorrespondingPTS[m_IDsAudio.m_nNumberFound] = i64CurrentPTS;
						m_IDsAudio.m_nNumberFound++;
						_i64toa (i64CurrentPTS, buffer, 16);
						sprintf (m_sOutBuffer, "%02d Audio Stream ID found: %02X at %s",m_IDsAudio.m_nNumberFound, i, buffer);
						Output (m_sOutBuffer);
						nLastStreamID = i;
					}
					else if (nStream == VIDEO && m_nRipType == RIP_TYPE_VIDEO 
						&& (i != nLastStreamID || i64abs (i64CurrentPTS, i64LastPTS) > 0x20000))
					{
						m_IDsVideo.m_nIDs[m_IDsVideo.m_nNumberFound] = i;
						m_IDsVideo.m_nBuffersSkipped[m_IDsVideo.m_nNumberFound] = nSkippedBuffers;
						m_IDsVideo.m_i64CorrespondingPTS[m_IDsVideo.m_nNumberFound] = i64CurrentPTS;
						m_IDsVideo.m_nNumberFound++;
						_i64toa (i64CurrentPTS, buffer, 16);
						sprintf (m_sOutBuffer, "%02d Video Stream ID found: %02X at %s",m_IDsVideo.m_nNumberFound, i, buffer);
						Output (m_sOutBuffer);
						nLastStreamID = i;
					}
					nSkippedBuffers += 100000;
					InStream.MoveBytes(100000);
					InStream.BitPointer(ui64BitPointer);
					i64LastPTS = i64CurrentPTS;
				}
				break;
			}
		}
		InStream.MoveBytes(1);
		nCount++;
		InStream.BitPointer(ui64BitPointer);
	}
	if (nLastStreamID == 0) return false;
	StreamID = nLastStreamID;
	return true;


}


bool CPVRFile::GetStreamIDs(CIBits &InStream, unsigned int &StreamID, int & nSkippedBuffers)
{
	int					i;
	unsigned __int64	ui64BitPointer;
	unsigned __int32	ui32Next4Bytes;
	
	StreamID=0;
	*pfsLog << "GetStreamIDs" << endl;

	*pfsLog << "try to rewind";
	if (!InStream.Rewind())
	{
		*pfsLog << "... failed..." << endl;
		return false;
	}

	*pfsLog << "... okay" << endl;
	Output ("Finding Stream ID");
	ui64BitPointer = nSkippedBuffers * 512 * 8;
  InStream.SeekTo (ui64BitPointer);
	while (NextStartCode(InStream, MBYTE * 10 * 10)) // && ui64BitPointer < (__int64) MBYTE * 80 * 10)
	{
#ifdef _WINPVR
		if (*m_pbCancel)
		{
			return false;
		}
#endif
		ui32Next4Bytes=InStream.Next4Bytes();
		//ui64BitPointer=InStream.NextBits(32);
		for (i=0xBD; i < 0xF0;i++)
		{
			switch(GetStreamType(i))
			{
			case OTHER:
				break;
			default:
				if (IsPES_Packet(InStream,i))
				{
          if ((m_nRipType == RIP_TYPE_VIDEO && GetStreamType(i)==VIDEO) ||
            (m_nRipType == RIP_TYPE_AUDIO && GetStreamType(i)!=VIDEO))
          {
					  StreamID=i;
					  sprintf (m_sOutBuffer, "Stream ID found: %02X", i);
					  return true;
          }
				}
				break;
			}
		}
		InStream.MoveBytes(1);
		InStream.BitPointer(ui64BitPointer);
	}
	Output ("Stream ID not found");
	nSkippedBuffers -= 2;
	if (nSkippedBuffers < 0) nSkippedBuffers = 0;
	return false;
}


bool CPVRFile::GetFirstPTS(CIBits* pcbStream, unsigned int uiStreamID, __int64 &i64PTS)
{
	unsigned __int32	ui32StartCode = 0x100 + uiStreamID;
	unsigned __int32	ui32Next4Bytes;
	unsigned __int64	i64SectorNumber;
	__int64				i64CurrentPTS;
	int					nBuffersToSkip, nLength;
	bool				bFound (false);
	// PES Packet
	CPES_Packet*		pppPES=NULL;
	// Streamtype
	STREAM_TYPE			uiStreamType=GetStreamType(uiStreamID);

	*pfsLog << setiosflags(ios::fixed) << resetiosflags(ios::dec) << setiosflags(ios::hex); 
	*pfsLog << endl << "FirstPTS " ;

	switch (uiStreamType)
	{
	case VIDEO:
		*pfsLog << "VIDEO ";
		nBuffersToSkip = m_nSkippedVideoBuffers;
		break;
	case AUDIO:
		*pfsLog << "AUDIO ";
		nBuffersToSkip = m_nSkippedAudioBuffers;
		break;
	case PRIVATE_1:
		*pfsLog << "PRIVATE_1 " ;
		nBuffersToSkip = m_nSkippedAudioBuffers;
		break;
	case PRIVATE_2:
		*pfsLog << "PRIVATE_2 " ;
		nBuffersToSkip = m_nSkippedAudioBuffers;
		break;
	}
	*pfsLog << (int)ui32StartCode << endl;

	*pfsLog << "try to rewind";
	if (!pcbStream->Rewind())
	{
		*pfsLog << "...failed returning" << endl;
		return false;
	}
	*pfsLog << "...okay" << endl;
  i64SectorNumber = m_nFirstClusterOffset * 512 * 8;
	pcbStream->SeekTo (i64SectorNumber);
	sprintf (m_sOutBuffer, "Finding PTS.... skipped %d Bytes", m_nFirstClusterOffset * 512 * 8);
	Output (m_sOutBuffer);
	while (true)
	{
#ifdef _WINPVR
		if (*m_pbCancel)
		{
			return false;
		}
#endif
		if (pcbStream->EOBS())
		{
			*pfsLog << "could not find any PTS that suites... return" << endl;
			return false;
		}
		if (pcbStream->Next4Bytes() == ui32StartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
			i64CurrentPTS=pppPES->PTS();

			if (pppPES->IsOkay() && i64CurrentPTS)
			{
				if (uiStreamType == VIDEO)
				{
					ui32Next4Bytes=pcbStream->Next4Bytes();
					if ((ui32Next4Bytes == 0x00000100 || ui32Next4Bytes == 0x000001B3) && pppPES->DTS_Available())
					{
						i64CurrentPTS = pppPES->PTS();
						nLength = pppPES->PES_Packet_Length ();
						bFound = true;
						*pfsLog << "found PTS " << (double)i64PTS << endl;
						*pfsLog << " waiting... for 0x000001B3" << endl;
					}
					delete pppPES;
				}
				else
				{
					delete pppPES;
					if (i64abs (i64PTS, i64CurrentPTS) < 0x80000 || i64PTS == -1)
					{
						i64PTS = i64CurrentPTS;
						*pfsLog << "found PTS " << (double)i64PTS << endl;
						*pfsLog << " audio pts... means doodeldee... gotit!!! -> return" << endl;
						return true;
					}
				}
			}
			else
			{
				delete pppPES;
			}
		}
		if (bFound && pcbStream->Next4Bytes() == 0x000001B3 )
		{
			if (i64abs (i64PTS, i64CurrentPTS) < 0x100000 || i64PTS == -1)
			{
				i64PTS = i64CurrentPTS;
				sprintf (m_sOutBuffer, "got 0x000001B3 after PTS %07I64X ... return\n", (unsigned __int64) i64PTS);
				Output (m_sOutBuffer);
				return true;
			}
			else 
			{
				bFound = false;
			}
		}
		else
		{
			pcbStream->MoveByte();
		}
	}
	// you may never be here...
	return false;
}

void CPVRFile::RipPrivateStream(CIBits* pcbStream, 
							 COBits* pcoOutput, 
							 unsigned int uiStreamID, 
							 __int64 i64StartPTS , 
							 __int64 i64EndPTS)
{
	unsigned __int32	ui32PESStartCode = 0x100 + uiStreamID;
	CPES_Packet*		pppPES=NULL;
	unsigned __int64	pos=0;
	__int64	LastPTS (0), i64CurrentPTS (0);
#ifdef _WINPVR
	unsigned __int32    i32StartPTS (0);
	unsigned __int32    i32EndPTS (0);
	unsigned __int32    i (0);
	unsigned __int64    i64Test (i64EndPTS);
#endif
	
	Output ("CPVRFile::RipPrivateStream\n");
//	*pfsLog << "CPVRFile::RipPrivateStream" << endl;
//	cout << "CPVRFile::RipPrivateStream" << endl;

	Output ("Try to rewind...");
//	*pfsLog << "Try to rewind...";
//	cout << "Try to rewind...";
	// first move to StartPTS
	if (!pcbStream->Rewind())
	{
		Output (" failed!\n");
		//*pfsLog << " failed!" << endl;
		//cout << " failed!" << endl;
		return;
	}

	Output (" passed!\n");
//	*pfsLog	<< " passed!" << endl;
//	cout << " passed!" << endl;
#ifdef _WINPVR
	if (*m_pbCancel) return;
	i32EndPTS =   (__int32) (i64EndPTS >> 3);
	i32StartPTS = (__int32) (i64StartPTS >> 3);
	m_pInfoAudio->m_nStartPos = i32StartPTS;
	m_pInfoAudio->m_nEndPos = i32EndPTS;
	m_pInfoAudio->m_nPosition = i32StartPTS;
	m_pInfoAudio->m_bRunning = true;
	m_pInfoAudio->m_bFirstUpdate = true;
	m_nRipType = RIP_TYPE_AUDIO;
	m_pInfoAudio->m_sInfo[0] = 0;
	i64Test = 0;
	m_bOverrun = false;
#endif
	SearchForStart (i64StartPTS, pcbStream, uiStreamID, m_sfdFile.iAudioFatEntry, m_sfdFile.iAudioClusterCount);
	while(!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
#ifdef _WINPVR
			if (*m_pbCancel)
			{
				delete pppPES;
				return;
			}
#endif
			// berlauf abfangen
			i64CurrentPTS = pppPES->PTS();
		  if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		  {
			  i64CurrentPTS += MAX_PTS_33BIT;
			  m_bOverrun = true;
		  }

			if (pppPES->IsOkay() && ((__int64)i64CurrentPTS >= i64StartPTS || i64abs (i64CurrentPTS, i64StartPTS) < 12000))
			{
				i64StartPTS = (__int64) i64CurrentPTS;
				LastPTS=i64CurrentPTS;
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
					// found this and next
					// 0.5.4
					pcbStream->SeekTo(pos);
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3,pcbStream,pcoOutput);
#ifdef _WINPVR		
					// Grsse bestimmen
					i64Test += pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3;
#endif
					sprintf (m_sOutBuffer, "found PTS %07I64X  running...\n", i64CurrentPTS);
					Output (m_sOutBuffer);
				//	*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
				//	cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
					delete pppPES;
					break;
				}
				else
				{
					*pfsLog << "this is not a PES_PACKET" << endl;
					delete pppPES;
					// 0.5.4
					pcbStream->SeekTo(pos+8);
					//pcbStream->MoveByte();
					NextStartCode(*pcbStream,MBYTE*10);
				}
			}
			else
			{
				*pfsLog << "PES ist not okay OR PTS is less then StartPTS" << endl;
				delete pppPES;
				pcbStream->MoveByte();
				NextStartCode(*pcbStream,MBYTE*10);
			}
		}
		else
		{
			*pfsLog << "this is not the expected startcode!" << endl;
			pcbStream->MoveByte();
			NextStartCode(*pcbStream,MBYTE*10);
		}
	}
	
	bool bBitStreamError=false;
	
	Output ("patience... running...\n", true, false);
	//cout << "patience... running..." << endl;
	// now we are here with valid data to process
	m_bOverrun = false;
	while (!pcbStream->EOBS())
	{
		pppPES = new CPES_Packet(pcbStream, uiStreamID);

#ifdef _WINPVR
		// Fortschrittsbalken anzeigen und Abbrucherkennung
    	m_pInfoAudio->m_nPosition = (__int32) (i64CurrentPTS >> 3);
		m_pInfoAudio->m_nFileSize = (__int32) (i64Test / (1024*1024));
		if (*m_pbCancel)
		{
			delete pppPES;
			return;
		}
#endif
		if (pppPES->IsOkay())
		{
			i64CurrentPTS = pppPES->PTS();
			// berlauf abfangen
		  if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		  {
			  i64CurrentPTS += MAX_PTS_33BIT;
			  m_bOverrun = true;
		  }

			if (i64CurrentPTS >= i64EndPTS)
			{
				delete pppPES;
				break;
			}
			else
			{
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
/*					if (i64CurrentPTS + 180000 < (__int64) LastPTS)
					{
						i64EndPTS = i64CurrentPTS + (i64EndPTS - i64StartPTS - (LastPTS - i64StartPTS));

						Output ("PTS overflow...\n");
						sprintf (m_sOutBuffer, "MAX PTS= %i64d\n", i64EndPTS);
						Output (m_sOutBuffer);
						//*pfsLog << "PTS overflow..." << endl;
						//*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						// cout << endl << "PTS overflow..." << endl;
						// cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						//it's not possible to know how far we're gone without timing information
						bBitStreamError=false;
					}
*/					// recovery!!!
					if (bBitStreamError)
					{
						sprintf (m_sOutBuffer, "DIFF= %I64d\n",i64CurrentPTS-LastPTS);
						Output (m_sOutBuffer);

						// *pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						// cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						AC3Recovery((long) (i64CurrentPTS-LastPTS),pcoOutput);
						bBitStreamError=false;
					}
					pcbStream->SeekTo(pos);
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3,pcbStream,pcoOutput);
#ifdef _WINPVR		
					// Grsse bestimmen
					i64Test += pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3;
#endif
					LastPTS=i64CurrentPTS;
				}
				else
				// if not PES then skip these bytes
				{
					bBitStreamError=true;
					Output("error recovery\n");
					// *pfsLog << "error recovery" << endl;
					// cout << "error recovery" << endl;
					pcbStream->SeekTo(pos+8);
					NextStartCode(*pcbStream);
				}
			}
		}
		else
			// if not PES then skip these bytes
		{
			bBitStreamError=true;
			Output("error recovery\n");
			// *pfsLog << "error recovery" << endl;
			// cout << "error recovery" << endl;
			// pcbStream->MoveByte();
			NextStartCode(*pcbStream);
		}
		delete pppPES;
	}

	if(pcbStream->EOBS())
	{
		Output ("EOBS reached!\n");
		// *pfsLog << "EOBS reached!" << endl;
		// cout << "EOBS reached!" << endl;
	}
}
void CPVRFile::RipAudioStream(CIBits* pcbStream, 
							 COBits* pcoOutput, 
							 unsigned int uiStreamID, 
							 __int64 i64StartPTS , 
							 __int64 i64EndPTS)
{
	unsigned __int32	ui32PESStartCode = 0x100 + uiStreamID;
	CPES_Packet*		pppPES=NULL;
	unsigned __int64	pos=0;
	__int64	LastPTS (0), i64CurrentPTS (0);
#ifdef _WINPVR
	unsigned __int32    i32StartPTS (0);
	unsigned __int32    i32EndPTS (0);
	unsigned __int32    i (0);
	unsigned __int64    i64Test (i64EndPTS);
#endif
	
	

	Output ("CPVRFile::RipAudioStream\n");
	// *pfsLog << "CPVRFile::RipAudioStream" << endl;
	// cout << "CPVRFile::RipAudioStream" << endl;

	Output ("Try to rewind...");
	// *pfsLog << "Try to rewind...";
	// cout << "Try to rewind...";
	// first move to StartPTS
	if (m_pInfoAudio->m_bRipRaw) // Rohdaten rippen
	{
		RipAudioStreamRaw (i64StartPTS, i64EndPTS, pcbStream, uiStreamID);
		return;
	}

	if (!pcbStream->Rewind())
	{
		Output ("failed!\n");
		// *pfsLog << " failed!" << endl;
		// cout << " failed!" << endl;
		return;
	}
	Output (" passed!\n");
	// *pfsLog	<< " passed!" << endl;
	// cout << " passed!" << endl;

#ifdef _WINPVR
	if (*m_pbCancel) return;
	while (i64Test > 0x1000000)
	{
		i64Test >>= 1;
		i++;
	}
	i32EndPTS = (__int32) i64Test;
	i32StartPTS = (__int32) (i64StartPTS >> i);
	m_pInfoAudio->m_nEndPos = i32EndPTS;
	m_pInfoAudio->m_nStartPos = i32StartPTS;
	m_pInfoAudio->m_nPosition = i32StartPTS;
	m_nRipType = RIP_TYPE_AUDIO;
	m_pInfoAudio->m_bRunning = true;
	m_pInfoAudio->m_bFirstUpdate = true;
	m_pInfoAudio->m_sInfo[0] = 0;
	i64Test = 0;
#endif

	SearchForStart (i64StartPTS, pcbStream, uiStreamID, m_sfdFile.iAudioFatEntry, m_sfdFile.iAudioClusterCount);
	m_bOverrun = false;
	while(!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
#ifdef _WINPVR
			if (*m_pbCancel)
			{
				delete pppPES;
				return;
			}
#endif
			// berlauf abfangen
			i64CurrentPTS = pppPES->PTS();
		  if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		  {
			  i64CurrentPTS += MAX_PTS_33BIT;
			  m_bOverrun = true;
		  }

			if (pppPES->IsOkay() && ((__int64) i64CurrentPTS >= i64StartPTS || i64abs (i64CurrentPTS, i64StartPTS) < 12000))
			{
				if (!m_bStreamWasCut || (i64abs (i64CurrentPTS, i64StartPTS) < 0x40000))
				{
					i64StartPTS = i64CurrentPTS;
					LastPTS=i64CurrentPTS;

					pcbStream->BitPointer(pos);
					pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
					if (IsPES_Packet(*pcbStream,uiStreamID))
					{
            bool bFoundMPEGHeader (false);
            unsigned __int32 ByteBuffer;
            unsigned __int64 i64PosOld;
            int i (0);

						// found this and next
						// 0.5.4
						pcbStream->SeekTo(pos);
  					pcbStream->BitPointer(i64PosOld);
            while (!bFoundMPEGHeader)
            {
           	  ByteBuffer=pcbStream->GetBits (32);
              if ((ByteBuffer & 0xFFFE) == 0xFFFC)
              {
      					pcbStream->BitPointer(pos);
                pcbStream->SeekTo(pos-16);
                bFoundMPEGHeader = true;
              }
              else
              {
                if (((ByteBuffer >> 16) & 0xFFFE) == 0xFFFC)
                {
      					  pcbStream->BitPointer(pos);
                  pcbStream->SeekTo(pos-32);
                  bFoundMPEGHeader = true;
                }
              }
            }
            pcbStream->BitPointer(pos);
            pos = (pos - i64PosOld) / 8;
						CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3 - (WORD) pos,pcbStream,pcoOutput);
						sprintf (m_sOutBuffer, "found PTS %i64d  running...\n", i64CurrentPTS);
						// *pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
						// cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "found PTS " << (double)pppPES->PTS() << " running..." << endl;
						delete pppPES;
						break;
					}
					else
					{
						// 0.5.4
						*pfsLog << "this is not a PES_PACKET" << endl;
						delete pppPES;
						//pcbStream->MoveByte();
						pcbStream->SeekTo(pos+8);
						NextStartCode(*pcbStream,MBYTE*10);
					}
				}
			}
			else
			{
				*pfsLog << "PES ist not okay OR PTS is less then StartPTS" << endl;
				delete pppPES;
				pcbStream->MoveByte();
				NextStartCode(*pcbStream,MBYTE*10);
			}
		}
		else
		{
			*pfsLog << resetiosflags(ios::dec) << setiosflags(ios::hex) << (long) pcbStream->Next4Bytes() << " is not the expected startcode!" << endl;
			pcbStream->MoveByte();
			NextStartCode(*pcbStream,MBYTE*10);
		}
	}

	
	bool bBitStreamError=false;
	Output ("patience... running...\n", true, false);
	// cout << "patience... running..." << endl;
	// now we are here with valid data to process
	while (!pcbStream->EOBS())
	{
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
#ifdef _WINPVR
		// Fortschrittsbalken anzeigen und Abbrucherkennung
    	m_pInfoAudio->m_nPosition = (__int32) (i64CurrentPTS >> i);
		m_pInfoAudio->m_nFileSize = (__int32) (i64Test / (1024*1024));
		if (*m_pbCancel)
		{
			delete pppPES;
			return;
		}
#endif
		if (pppPES->IsOkay())
		{
			i64CurrentPTS = pppPES->PTS();
			// berlauf abfangen
		  if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		  {
			  i64CurrentPTS += MAX_PTS_33BIT;
			  m_bOverrun = true;
		  }
			if ((__int64) i64CurrentPTS >= i64EndPTS || m_i64AudioMaxPos < pcbStream->BitPointer())
			{
				delete pppPES;
				break;
			}
			else
			{
				pcbStream->BitPointer(pos);
				pcbStream->SeekTo(pos + (pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3)*8);
				if (IsPES_Packet(*pcbStream,uiStreamID))
				{
/*					if (pppPES->PTS() + 180000 < (__int64) LastPTS)
					{
						i64EndPTS = pppPES->PTS() + (i64EndPTS - i64StartPTS - (LastPTS - i64StartPTS));
						Output ("PTS overflow...\n");
						sprintf (m_sOutBuffer, "Max PTS= %i64d\n", i64EndPTS);
						Output (m_sOutBuffer);

						// *pfsLog << "PTS overflow..." << endl;
						// *pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						// cout << endl << "PTS overflow..." << endl;
						// cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "Max PTS=" << double(i64EndPTS) << endl;
						//it's not possible to know how far we're gone without timing information
						bBitStreamError=false;
					}
*/
					// recovery!!!
					if (bBitStreamError)
					{	
						sprintf (m_sOutBuffer, "DIFF=%d\n", (int)(i64CurrentPTS-LastPTS));
						// *pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						// cout << resetiosflags(ios::hex) << setiosflags(ios::dec) << "DIFF=" << (int)(pppPES->PTS()-LastPTS) << endl;
						MPEG_Audio_Recovery((long) (i64CurrentPTS-LastPTS),pcoOutput);
						bBitStreamError=false;
					}
					pcbStream->SeekTo(pos);
          if ( ((pcbStream->Next4Bytes ()) >> 16) == 0xFFFF)
          {
            pcbStream->MoveBytes (2);
            pos = 2;
          }
          else
          {
            pos = 0;
          }
					CopyBytes(pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3 - (WORD) pos,pcbStream,pcoOutput);
#ifdef _WINPVR		
					// Grsse bestimmen
					i64Test += pppPES->PES_Packet_Length() - pppPES->PES_Header_Data_Length() - 3 - pos;
#endif
					LastPTS=i64CurrentPTS;
				}
				else
				// if not PES then skip these bytes
				{
					bBitStreamError=true;
					Output ("error recovery\n");
					// *pfsLog << "error recovery" << endl;
					// cout << "error recovery" << endl;
					pcbStream->SeekTo(pos+8);
					NextStartCode(*pcbStream);
				}

			}
		}
		else
			// if not PES then skip these bytes
		{
			bBitStreamError=true;
			Output ("error recovery\n");
			// *pfsLog << "error recovery" << endl;
			// cout << "error recovery" << endl;
			pcbStream->MoveByte();
			NextStartCode(*pcbStream);
		}
		delete pppPES;
	}

	if(pcbStream->EOBS())
	{
		Output ("EOBS reached!\n");
		// *pfsLog << "EOBS reached!" << endl;
		// cout << "EOBS reached!" << endl;
	}
}

void CPVRFile::CopyBytes(unsigned int BytesToCopy, CIBits* pcbStream, COBits* pcoOutput)
{
	unsigned int iByte;
	m_RecoveryBuffer.erase(m_RecoveryBuffer.begin(),m_RecoveryBuffer.end());
	while (!pcbStream->EOBS() && BytesToCopy--)
	{
		pcbStream->GetByte(iByte);
		pcoOutput->PutByte(iByte);
		m_RecoveryBuffer.push_back(iByte);
	}
}

void CPVRFile::AC3Recovery(long PTSDiff, COBits* pcoOutput)
{
//	for (int i=0; i < m_RecoveryBuffer.size(); i++)
	double iFramesPerS=448000.0 / 8 / 1792; //31,25
	int iFrames = (int) ((double) PTSDiff / 90000.0 * iFramesPerS + 0.5);

	int iBytes=iFrames*1792;
	int i=0;
	while (iBytes--)
	{
		pcoOutput->PutByte(m_RecoveryBuffer[i++]);
		if (i>=(int) m_RecoveryBuffer.size())
			i=0;
	}
}

void CPVRFile::MPEG_Audio_Recovery(long PTSDiff, COBits* pcoOutput)
{
//	for (int i=0; i < m_RecoveryBuffer.size(); i++)
	double iFramesPerS=192000.0 / 8 / 576; //31,25
	int iFrames = (int) ((double)PTSDiff / 90000 * iFramesPerS + 0.5);

	int iBytes=iFrames*576;
	int i=0;
	while (iBytes--)
	{
		pcoOutput->PutByte(m_RecoveryBuffer[i++]);
		if (i>= (int) m_RecoveryBuffer.size())
			i=0;
	}
}

void CPVRFile::RipVideoStream(CIBits* pcbStream, 
						   COBits* pcoOutput, 
						   unsigned int uiStreamID, 
						   __int64 i64StartPTS , 
						   __int64 i64EndPTS)
{
	unsigned __int32	ui32PESStartCode = 0x100 + uiStreamID;
	CPES_Packet*		pppPES (NULL);
	unsigned int		uiByte (0);
	bool				PTS_Reached (false), bEndReached (false), bDelayWritten (false);
	STREAM_TYPE			stStreamType=GetStreamType(uiStreamID);
	__int64				i64LastPTS=0, i64ThisPTS=0, i64CurrentPTS (0), i64RealStartPTS (0);
#ifdef _WINPVR
	unsigned __int32    i32StartPTS (0);
	unsigned __int32    i32EndPTS (0);
	unsigned __int32    i (0), nCount (0);
	unsigned __int64    i64Test (i64EndPTS);
	CStdioFile			fDelayOut;
  CString sFilename;
 	unsigned __int32 ByteBuffer;

  __int64 i64PicturePTS (0), i64Diff (0), i64PicDiff (0);

#endif


	// better be good to me
	i64EndPTS--;


	Output ("CPVRFile::RipVideoStream\n");
	// *pfsLog << endl << "CPVRFile::RipVideoStream" << endl;
	// cout << endl << "CPVRFile::RipVideoStream" << endl;
	cout << setiosflags(ios::fixed) ;
	
	Output ("rewinding stream");
	// *pfsLog << "rewinding stream";
	// cout << "rewinding stream";
	if (m_pInfoVideo->m_bRipRaw) // Rohdaten rippen
	{
		RipVideoStreamRaw (i64StartPTS, i64EndPTS, pcbStream, uiStreamID);
		return;
	}

	if (pcbStream->Rewind())
	{
		Output ("...okay\n");
		// *pfsLog << "...okay";
		// cout << "...okay";
	}
	else
	{
		Output ("...failed!!!\n");
		// *pfsLog << "...failed!!!";
		// cout << "...failed!!!";
	}
	// *pfsLog << endl;
	// cout << endl;

#ifdef _WINPVR
	if (*m_pbCancel) return;
	while (i64Test > 0x10000000)
	{
		i64Test >>= 1;
		i++;
	}
	i32EndPTS = (__int32) i64Test;
	i32StartPTS = (__int32) (i64StartPTS >> i);

	m_pInfoVideo->m_nEndPos = i32EndPTS;
	m_pInfoVideo->m_nStartPos = i32StartPTS;
	m_pInfoVideo->m_nPosition = i32StartPTS;
	m_pInfoVideo->m_i64StartPos = i64StartPTS;
	m_pInfoVideo->m_i64EndPos = i64EndPTS;
	m_pInfoVideo->m_bRunning = true;
	m_pInfoVideo->m_bFirstUpdate = true;
	m_nRipType = RIP_TYPE_VIDEO;
	m_pInfoVideo->m_sInfo[0] = 0;
	i64Test = 0;
#endif
	sprintf (m_sOutBuffer, "seeking stream start code %X\n", (unsigned int)ui32PESStartCode);
	Output (m_sOutBuffer);
	// *pfsLog << resetiosflags(ios::dec) << setiosflags(ios::hex);
	// *pfsLog << "seeking stream start code " << (unsigned int)ui32PESStartCode << endl; 
	// cout << resetiosflags(ios::dec) << setiosflags(ios::hex);
	// cout << "seeking stream start code " << (unsigned int)ui32PESStartCode << endl;
	SearchForStart (i64StartPTS, pcbStream, uiStreamID, m_sfdFile.iVideoFatEntry, m_sfdFile.iVideoClusterCount);
	m_bOverrun = false;
	while (!pcbStream->EOBS())
	{
		if (pcbStream->Next4Bytes() == ui32PESStartCode)
		{
			if (!PTS_Reached)
			{
				pppPES = new CPES_Packet(pcbStream, uiStreamID);
#ifdef _WINPVR
				if (*m_pbCancel)
				{
					delete pppPES;
					return;
				}
#endif
				i64LastPTS = i64CurrentPTS;
				// berlauf abfangen
				i64CurrentPTS = pppPES->PTS();
				if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
				{
					i64CurrentPTS += MAX_PTS_33BIT;
					m_bOverrun = true;
				}

				if (pppPES->IsOkay() && (i64CurrentPTS >= i64StartPTS || i64abs (i64CurrentPTS, i64StartPTS) < 25))
				{
					if (!m_bStreamWasCut || (i64abs (i64CurrentPTS, i64StartPTS) < (__int64) 0x1000000))
					{
		        sprintf (m_sOutBuffer, "reached pts %07I64X\n", i64CurrentPTS);
		        Output (m_sOutBuffer);
						PTS_Reached=true;
            Output ("shit.. i'm video and i have to wait until 0x000001B3....\n");
					}
				}
				delete pppPES;
			}
		}
		if (PTS_Reached && pcbStream->NextBits(32)==0x000001B3) // Aspect Ratio Bestimmung
		{
      ByteBuffer=pcbStream->GetBits(32);      
    	ByteBuffer=pcbStream->GetBits(12);
    	ByteBuffer=pcbStream->GetBits(12);
    	ByteBuffer=pcbStream->GetBits(4);
      m_dAspectRatio = 4.0 / 3.0;
      if (ByteBuffer == 3) m_dAspectRatio = 16.0 / 9.0;
      pcbStream->SeekTo(pcbStream->BitPointer()-(32+12+12+4));
			*pfsLog << "yeah.. i'm video and i reached 0x000001B3...." ;
      if (ByteBuffer == 3) 
      {
        m_dAspectRatio = 16.0 / 9.0;
        *pfsLog << " aspect ratio = 16:9 " << endl;
      }
      else
      {
        m_dAspectRatio = 4.0 / 3.0;
        *pfsLog << " aspect ratio = 4:3 " << endl;
      }

			break;
		} // Aspect Ratio
		pcbStream->MoveByte();
		NextStartCode(*pcbStream);
	}
  
  m_i64CurrPos = i64CurrentPTS;
  i64RealStartPTS = i64CurrentPTS;
	i64LastPTS=i64CurrentPTS;
/*  ByteBuffer=pcbStream->GetBits(32);      
  while (ByteBuffer != 0x000001B8)  // 1. GOP suchen
  {
    pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
	  ByteBuffer = ((ByteBuffer & 0x00FFFFFF) << 8) | pcbStream->GetByte();
  }
  if (ByteBuffer == 0x000001B8)
  {
    // Timing Info eintragen
		pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
		pcoOutput->PutByte((ByteBuffer & 0xFF0000) >> 16);
		pcoOutput->PutByte((ByteBuffer & 0xFF00) >> 8);
		pcoOutput->PutByte((ByteBuffer & 0xFF));
    __int64 i64Diff;
    int hour, min, sec, frame;
    min = pcbStream->GetBits (1); // Drop Bit
    pcoOutput->PutBits (min, 1);
    pcbStream->GetBits (24); 
    i64Diff = i64CurrentPTS - i64RealStartPTS;
    hour = (int) (i64Diff / (90000 * 3600));
    if (hour > 23) hour = 0;
    pcoOutput->PutBits (hour, 5);
    min = (int) (i64Diff / (90000 * 60)) % 60;
    pcoOutput->PutBits (min, 6);
    min = 1;
    pcoOutput->PutBits (min, 1);
    sec = (int) (i64Diff / 90000) % 60;
    pcoOutput->PutBits (sec, 6);
    frame = (int) (i64Diff / (90000/25) + 1) % 25;
    pcoOutput->PutBits (frame, 6);
    min = pcbStream->GetBits (7); 
    pcoOutput->PutBits (min, 7);
    ByteBuffer=pcbStream->GetBits(32);
#ifdef _WINPVR
		i64Test+=8;
#endif
  }
  pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
	ByteBuffer = ((ByteBuffer & 0x00FFFFFF) << 8) | pcbStream->GetByte();
  while (ByteBuffer != 0x000001B8)
  {
    if (ByteBuffer == 0x00000100) // Bildkennung
    {
      ByteBuffer = pcbStream->GetBits(10);
      ByteBuffer = pcbStream->GetBits(3);
      if (ByteBuffer == B_FRAME) // B-Frame muss raus
      {
        ByteBuffer = pcbStream->NextBits(3);
        while (ByteBuffer != 0x00000100)  // nchstes Bild
        {
      	  ByteBuffer = ((ByteBuffer & 0x00FFFFFF) << 8) | pcbStream->GetByte();
        }
      }
      else
      {
        pcbStream->SeekTo(pcbStream->BitPointer()-(10+3));
      }
      ByteBuffer = 0x00000100;
    }
    pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
	  ByteBuffer = ((ByteBuffer & 0x00FFFFFF) << 8) | pcbStream->GetByte();
  }
  */
	if (pcbStream->EOBS())
	{
		sprintf (m_sOutBuffer, "EOF reached! No PTS for Code %X\n", (unsigned int)ui32PESStartCode);
		Output (m_sOutBuffer);
		//*pfsLog << "EOF reached! No PTS for Code " << (unsigned int)ui32PESStartCode << endl;
		// cout << "EOF reached! No PTS for Code " << (unsigned int)ui32PESStartCode << endl;
	}

	*pfsLog << resetiosflags(ios::hex) << setiosflags(ios::dec);
	cout << resetiosflags(ios::hex) << setiosflags(ios::dec);
	
	Output ("now ripping until eobs...\n");
	// cout << "now ripping until eobs..." << endl;
	// *pfsLog << "now ripping until eobs..." << endl;
	cout << setiosflags(ios::fixed);

//  ByteBuffer=pcbStream->GetBits(32);
	while (!pcbStream->EOBS())
	{
		if (ByteBuffer == ui32PESStartCode)
		{

			pcbStream->SeekTo(pcbStream->BitPointer()-32);
			pppPES = new CPES_Packet(pcbStream, uiStreamID);
			if (pppPES->IsOkay())
			{
#ifdef _WINPVR
				// Fortschrittsbalken anzeigen und Abbrucherkennung
				if (nCount++ > 10)
				{
    				m_pInfoVideo->m_nPosition = (__int32) (i64ThisPTS >> i);
					m_pInfoVideo->m_nFileSize = (__int32) (i64Test / (1024*1024));
					nCount = 0;
				}
				if (*m_pbCancel)
				{
					delete pppPES;
					return;
				}

#endif
				// berlauf abfangen
				i64CurrentPTS = pppPES->PTS();
        if (!bDelayWritten)
        {
          sFilename = Filename ().c_str ();
          sFilename += " delay.log";
  			  if (fDelayOut.Open (sFilename, CFile::modeCreate | CFile::modeWrite))
				  {
	  			  sprintf (m_sOutBuffer, "Delay: %.0f ms", (double) (i64CurrentPTS - i64StartPTS) / 90.0);
					  fDelayOut.WriteString (m_sOutBuffer);
					  fDelayOut.Close ();
				  }
          bDelayWritten = true;
        }

				if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
				{
					i64CurrentPTS += MAX_PTS_33BIT;
					m_bOverrun = true;
				}
				
				i64ThisPTS=i64CurrentPTS;

				if (pppPES->DTS_Available())
				{
#ifndef _WINPVR
					__int32 nCount;
					nCount++;
					if (nCount > 100)
					{
						nCount = 0;
#endif
          __int64 i64Temp;
            i64Temp = i64ThisPTS - m_i64FirstPTS;
						sprintf (m_sOutBuffer, " at PTS = %d:%02d:%02d\n", (int)(i64Temp / (90000 *3600)), (int)(i64Temp / (90000 *60)) % 60,(int)(i64Temp / 90000) % 60);
						Output (m_sOutBuffer, true, false);
						// cout << " at PTS=" << (double)i64ThisPTS << (char)13;
#ifndef _WINPVR
					}
#endif
				}
				if (i64ThisPTS > i64EndPTS && i64abs (i64EndPTS, i64ThisPTS) < (MAX_PTS_33BIT >> 2) || m_i64VideoMaxPos < pcbStream->BitPointer())
				{
					sprintf (m_sOutBuffer, " at PTS = %07I64X ", i64ThisPTS);
					Output (m_sOutBuffer, true, true);
					cout << (char)13 << endl;
					 // cout << " at PTS=" << (double)i64ThisPTS << (char)13 << endl;
					 // *pfsLog << " at PTS=" << (double)i64ThisPTS << endl;
					Output ("Stream ripped successfully...\n");
					// *pfsLog << "Stream ripped successfully..." << endl;
					// cout << "Stream ripped successfully..." << endl;
					delete pppPES;
					break;
				}
				i64LastPTS=i64ThisPTS;
				ByteBuffer=pcbStream->GetBits(32);
			}
			// not okay...
			else
			{
#ifdef _WINPVR
				if (*m_pbCancel)
				{
					delete pppPES;
					return;
				}
#endif
				//pcbStream->GetByte(uiByte);
				//pcoOutput->PutByte(uiByte);
				pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
				ByteBuffer = ((ByteBuffer & 0x00FFFFFF) << 8) | pcbStream->GetByte();
			}
			delete pppPES;
		}
		else
		{
			//pcoOutput->PutByte(pcbStream->GetByte());
      if (ByteBuffer == 0x000001B8)
      {
        // Timing Info eintragen
			  pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
			  pcoOutput->PutByte((ByteBuffer & 0xFF0000) >> 16);
			  pcoOutput->PutByte((ByteBuffer & 0xFF00) >> 8);
			  pcoOutput->PutByte((ByteBuffer & 0xFF));
        __int64 i64Diff;
        int hour, min, sec, frame;
        min = pcbStream->GetBits (1); // Drop Bit
        pcoOutput->PutBits (min, 1);
        pcbStream->GetBits (24); 
        i64Diff = i64CurrentPTS - i64RealStartPTS;
        hour = (int) (i64Diff / (90000 * 3600));
        if (hour > 23) hour = 0;
        pcoOutput->PutBits (hour, 5);
        min = (int) (i64Diff / (90000 * 60)) % 60;
        pcoOutput->PutBits (min, 6);
        min = 1;
        pcoOutput->PutBits (min, 1);
        sec = (int) (i64Diff / 90000) % 60;
        pcoOutput->PutBits (sec, 6);
        frame = (int) (i64Diff / (90000/25) + 1) % 25;
        pcoOutput->PutBits (frame, 6);
        min = pcbStream->GetBits (7); 
        pcoOutput->PutBits (min, 7);
        ByteBuffer=pcbStream->GetBits(32);
  #ifdef _WINPVR
			  i64Test+=8;
  #endif
      }
      else
      {
			  pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
  #ifdef _WINPVR
			  i64Test++;
  #endif
			  ByteBuffer = ((ByteBuffer & 0x00FFFFFF) << 8) | pcbStream->GetByte();
      }
		}
	}

	if (pcbStream->EOBS())
	{
		pcoOutput->PutByte((ByteBuffer & 0xFF000000) >> 24);
		pcoOutput->PutByte((ByteBuffer & 0x00FF0000) >> 16);
		pcoOutput->PutByte((ByteBuffer & 0x0000FF00) >> 8);
		pcoOutput->PutByte(ByteBuffer & 0x000000FF);
		Output ("EOBS Reached...\n");
		// *pfsLog << "EOBS Reached..." << endl;
		// cout << "EOBS Reached..." << endl;
	}
#ifdef _WINPVR
  	m_pInfoVideo->m_nPosition = i32EndPTS;
#endif
	pcoOutput->Flush();
	*pfsLog << "return" << endl;
}


void CPVRFile::RipStreams(double iStartSeconds,double iSeconds, bool bWriteLog)
{
	CIBits*					cbVideoStream=NULL;
	CIBits*					cbAudioStream=NULL;
	unsigned int			uiVideoStreamID=0;
	unsigned int			uiAudioStreamID=0;
  CFile           fMPEG2File;
  CString s;

	COBits*			coOut=NULL;
	__int64			i64FirstPTS=0, i64LastPTS=0;
	int				i (0), j (0);

	DWORD			tick=0,tack=0;

	bool			bResult=false;

	cout.precision(0);
	Dump(&cout,false);
	Dump(pfsLog,false);
	
	cout << "analyzing streams..." << endl;
  m_dOffsetTime = iStartSeconds;
  m_dRipTime = iSeconds;

	m_bStreamWasCut = false;
	m_pInfoAudio->m_sInfo[0] = 0;
	m_pInfoVideo->m_sInfo[0] = 0;
	m_nFirstClusterOffset = 0;
	if (m_pInfoVideo->m_bMakeBackup) // Backup der Blocks machen
	{
//		LookupClusters ();
		BackupClusters ();
		return;
	}
	if (m_pInfoAudio->m_bMakeBackup)
	{
//		RestoreClusters ();
//		return;
	}

#ifdef _WINPVR
	if (m_pInfoVideo->m_bProcess)
	{
#endif
	m_nRipType = RIP_TYPE_VIDEO;
	cbVideoStream=new CIBits(m_sDrivename,m_sfdFile.iVideoFatEntry);
	m_nSkippedVideoBuffers = m_sfdFile.iVideoStartFirstCluster;
  m_nLastClusterOffset = m_sfdFile.iVideoStopFirstCluster;
	GetStreamIDs (*cbVideoStream, uiVideoStreamID, m_nSkippedVideoBuffers);
  m_i64VideoMaxPos = GetMaxPos (m_sfdFile.iVideoStopFirstCluster, m_sfdFile.iVideoClusterCount);
	cbVideoStream->Rewind();
#ifdef _WINPVR
	}
#endif

#ifdef _WINPVR
	if (m_pInfoAudio->m_bProcess)
	{
#endif
	m_nRipType = RIP_TYPE_AUDIO;
	cbAudioStream=new CIBits(m_sDrivename,m_sfdFile.iAudioFatEntry);
	m_nSkippedAudioBuffers = m_sfdFile.iAudioStartFirstCluster;
  m_nLastClusterOffset = m_sfdFile.iAudioStopFirstCluster;
	GetStreamIDs(*cbAudioStream, uiAudioStreamID,  m_nSkippedAudioBuffers);
  m_i64AudioMaxPos = GetMaxPos (m_sfdFile.iAudioStopFirstCluster, m_sfdFile.iAudioClusterCount);
	cbAudioStream->Rewind();
#ifdef _WINPVR
	}
#endif
	cout << setiosflags(ios::hex) ;
	*pfsLog << setiosflags(ios::hex) ;

	*pfsLog << "Videostream=";
	cout << "Videostream=";

	cout << uiVideoStreamID << " ";
	*pfsLog << uiVideoStreamID << " ";

	cout << endl;
	*pfsLog << endl;

	cout << "Audiostream=";
	*pfsLog << "Audiostream=";


	cout << uiAudioStreamID << " ";
	*pfsLog << uiAudioStreamID << " ";
	cout << endl;
	*pfsLog << endl;

#ifdef _WINPVR
	if ((uiAudioStreamID || !m_pInfoAudio->m_bProcess) && 
		(uiVideoStreamID || !m_pInfoVideo->m_bProcess))
#else
	if (uiAudioStreamID && uiVideoStreamID)
#endif
	{

#ifdef _WINPVR
	if (m_pInfoAudio->m_bProcess)
	{
#endif
		m_nRipType = RIP_TYPE_AUDIO;
		cout << "try to find first pts for Audio..." << endl;
    m_nFirstClusterOffset = m_nSkippedAudioBuffers;
    m_nLastClusterOffset = m_sfdFile.iAudioStopFirstCluster;
    if (m_i64PTSAudio == -1)
		  GetFirstPTS(cbAudioStream, uiAudioStreamID, m_i64PTSAudio);
		sprintf (m_sOutBuffer, "AudioStream PTS=%07I64X\n", m_i64PTSAudio);
    Output (m_sOutBuffer);
#ifdef _WINPVR
	}
#endif

#ifdef _WINPVR
	if (m_pInfoVideo->m_bProcess)
	{
#endif
		m_nRipType = RIP_TYPE_VIDEO;
		cout << "try to find first pts for Video..." << endl;
    m_nFirstClusterOffset = m_nSkippedVideoBuffers;
    m_nLastClusterOffset = m_sfdFile.iVideoStopFirstCluster;
    if (m_i64PTSVideo == -1)
		  GetFirstPTS(cbVideoStream, uiVideoStreamID, m_i64PTSVideo);
		sprintf (m_sOutBuffer, "Videostream PTS=%07I64X\n", m_i64PTSVideo);
    Output (m_sOutBuffer);
#ifdef _WINPVR
	}
#endif
		if (!m_pInfoVideo->m_bProcess)
		{
			i64FirstPTS = m_i64PTSAudio + (unsigned __int64) (iStartSeconds * 90000);
      m_i64FirstPTS = m_i64PTSAudio;
		}
		else if (!m_pInfoAudio->m_bProcess)
		{
			i64FirstPTS = m_i64PTSVideo + (unsigned __int64) (iStartSeconds * 90000);
      m_i64FirstPTS = m_i64PTSVideo;
		}
		else
		{
			i64FirstPTS = m_i64PTSVideo + (unsigned __int64) (iStartSeconds * 90000);
      m_i64FirstPTS = m_i64PTSVideo;
		}
    m_pInfoVideo->m_i64ClipStart = m_i64FirstPTS;    
		i64LastPTS = i64FirstPTS + (unsigned __int64) (90000 * iSeconds);
		sprintf (m_sOutBuffer, "First PTS=%07I64X\nMax   PTS=%07I64X\n", i64FirstPTS,i64LastPTS);
    Output (m_sOutBuffer);

#ifdef _WINPVR
	if (m_pInfoAudio->m_bProcess)
	{
#endif
		m_nRipType = RIP_TYPE_AUDIO;
		cout << "now ripping audio" << endl;
		tick = GetTickCount();
    s = "." + m_pInfoAudio->m_sPrefix;
		if(GetStreamType(uiAudioStreamID)==AUDIO || m_pInfoAudio->m_bRipRaw)
		{
			if (!m_pInfoAudio->m_bRipRaw)
				coOut=new COBits(Filename().append(s), bResult);
			m_nFirstClusterOffset = m_nSkippedAudioBuffers;
      m_nLastClusterOffset = m_sfdFile.iAudioStopFirstCluster;

			RipAudioStream(cbAudioStream,coOut,uiAudioStreamID,i64FirstPTS,i64LastPTS);
		}
		else
		{
			coOut=new COBits(Filename().append(".ac3"), bResult);
			m_nFirstClusterOffset = m_nSkippedAudioBuffers;
      m_nLastClusterOffset = m_sfdFile.iAudioStopFirstCluster;
			RipPrivateStream(cbAudioStream,coOut,uiAudioStreamID,i64FirstPTS,i64LastPTS);
		}
		tack = GetTickCount()-tick;
		sprintf (m_sOutBuffer, "... it took %d  seconds...\n", (int)tack / 1000);
    Output (m_sOutBuffer);
		if (!m_pInfoAudio->m_bRipRaw) delete coOut;

#ifdef _WINPVR
		m_pInfoAudio->m_nRipTime = (int)tack / 1000;
	}
#endif

#ifdef _WINPVR
	if (m_pInfoVideo->m_bProcess) 
	{
#endif
		m_nRipType = RIP_TYPE_VIDEO;
		cout << "now ripping video" << endl;
		tick = GetTickCount();
    s = "." + m_pInfoVideo->m_sPrefix;
    if (!m_pInfoVideo->m_bRipRaw)	coOut=new COBits(Filename().append(s),bResult);
		m_nFirstClusterOffset = m_nSkippedVideoBuffers;
    m_nLastClusterOffset = m_sfdFile.iVideoStopFirstCluster;

		RipVideoStream(cbVideoStream,coOut,uiVideoStreamID,i64FirstPTS,i64LastPTS);
		// Sequence END Code
    if (!m_pInfoVideo->m_bRipRaw)
    {
	    coOut->PutByte(0x00);
	    coOut->PutByte(0x00);
	    coOut->PutByte(0x01);
	    coOut->PutByte(0xb7);
	    coOut->Flush();
      s = coOut->GetFilename ();
	    delete coOut;
      // Patch Sequence Header
      if (m_pInfoVideo->m_bPatchSeqHeader)
      {
        if (fMPEG2File.Open (s, CFile::modeReadWrite))
        {
          int nSeqCount (0), nSequPos (-1);
          DWORD nTime, nSize;
          int nRes;
          fMPEG2File.Read (m_pBuffer,MBYTE);
          for (i = 0; i < MBYTE && nSeqCount < 2; i++)
          {
            if (m_pBuffer[i] == 0 && m_pBuffer[i+1] == 0 && m_pBuffer[i+2] == 1 && m_pBuffer[i+3] == 0xB3)
            {
              if (nSeqCount == 0) nSequPos = i;
              nSeqCount++;
              nRes = (m_pBuffer[i+4] << 4) + (m_pBuffer[i+5] >> 4);
              if (nRes != 352 && nRes != 704 && nRes != 720 && nSeqCount < 2)
              {
                m_pBuffer[i+4] = (352 >> 4);
                m_pBuffer[i+5] &= 0xF;
                m_pBuffer[i+5] |= (BYTE) (352 << 4) && 0xF0;
              }
              nTime = (int) m_dRipTime;
              nSize = (DWORD) (fMPEG2File.GetLength () / nTime);
              nSize = (nSize * 8);
              nSize /= 400;
              m_pBuffer[i+8] = (BYTE) (nSize >> 10);
              m_pBuffer[i+9] = (BYTE) (nSize >> 2) & 0xFF;
              m_pBuffer[i+10] &= ~0xC0;
              m_pBuffer[i+10] |= (BYTE) (nSize << 6) & 0xC0;
            }
          }
          fMPEG2File.Seek (nSequPos, CFile::begin);
          fMPEG2File.Write (&m_pBuffer[nSequPos], MBYTE);
          fMPEG2File.Close ();
        }
      }
    }

		tack = GetTickCount()-tick;
		sprintf (m_sOutBuffer, "... it took %d  seconds...\n", (int)tack / 1000);
    Output (m_sOutBuffer);
#ifdef _WINPVR
		m_pInfoVideo->m_nRipTime = (int)tack / 1000;
	}
#endif
	}
	else
	{
		cout << "can't go on..." << endl;
		*pfsLog << "can't go on..." << endl;
	}
	cout << "cleaning up" << endl;

#ifdef _WINPVR
	if (m_pInfoVideo->m_bProcess) 
	{
#endif
	delete cbVideoStream;
#ifdef _WINPVR
	}
#endif

#ifdef _WINPVR
	if (m_pInfoAudio->m_bProcess) 
	{
#endif
	delete cbAudioStream;
#ifdef _WINPVR
	}
#endif

}

void CPVRFile::Dump(ostream *posOut, bool bWait)
{
  char string[255];
  sprintf (string, "%d",m_sfdFile.iRecordNo);
	*posOut << "RecordNo    = " << string << endl;
	*posOut << "Station     = " << m_sfdFile.cStationName << endl;
	*posOut << "Clip        = " << m_sfdFile.cClipName << endl;
  sprintf (string, "%02d.%02d.%02d - %d:%02d:%02d",m_sfdFile.iRecordDay, m_sfdFile.iRecordMonth, m_sfdFile.iRecordYear,
    m_sfdFile.iRecordHour, m_sfdFile.iRecordMinute, m_sfdFile.iRecordSecond);
  
  *posOut << "Date        = " 
		<< string
		<< endl;
  sprintf (string, "%d:%02d:%02d",m_sfdFile.iLengthHours ,m_sfdFile.iLengthMinutes,m_sfdFile.iLengthSeconds);

	*posOut << "Length      = " 
		<< string
		<< endl;
  sprintf (string, "%d",m_sfdFile.iLengthHours * 3600 + m_sfdFile.iLengthMinutes * 60 + m_sfdFile.iLengthSeconds);
	*posOut << "Length in s =" 
		<< string
		<< endl;
	*posOut << "SkipIndex   =" << endl;
	int i=0;
	while(i<MAX_SKIP_INDEX)
	{
		if (m_sfdFile.ssiSkipIndex[i].iStart!=-1 && m_sfdFile.ssiSkipIndex[i].iStop!=-1)
		{
      sprintf (string, "start=%d / stop=%d",m_sfdFile.ssiSkipIndex[i].iStart, m_sfdFile.ssiSkipIndex[i].iStop);
  		cout 
			<< string
			<< endl;
		}
		i++;
	}
	*posOut << "VClusters   = " 
		<< m_sfdFile.iVideoClusterCount
		<< " means max " 
		<< m_sfdFile.iVideoClusterCount * 100 
		<< " MByte"
		<< endl;
	*posOut << "AClusters   = " 
		<< m_sfdFile.iAudioClusterCount
		<< " means max " 
		<< m_sfdFile.iAudioClusterCount * 100 
		<< " MByte"
		<< endl;
	*posOut << "VClusters   =" ;
	g_pfatFAT->DumpClusters(posOut, m_sfdFile.iVideoFatEntry);
	*posOut << "AClusters   =" ;
	g_pfatFAT->DumpClusters(posOut, m_sfdFile.iAudioFatEntry);
	*posOut << endl;
}


void CPVRFile::Output(char *s, bool bOutConsole, bool bOutFile)
{
	bool bNewline (false);
	char buf[255];
	strcpy (buf, s);

	if (buf[strlen(buf)-1] == '\n') 
	{
		buf[strlen(buf) - 1] = 0;
		bNewline = true;
	}
#ifdef _WINPVR
	if (m_nRipType == RIP_TYPE_AUDIO)
		strcpy (m_pInfoAudio->m_sInfo, buf);
	else
		strcpy (m_pInfoVideo->m_sInfo, buf);
#else
	if (bOutConsole) 
	{
		cout << buf;
		if (bNewline) cout << endl;
	}
#endif
	if (bOutFile)
	{
		*pfsLog << buf;
		if (bNewline) *pfsLog << endl;
	}
}

// Schnittstelle zur GUI
#ifdef _WINPVR
sFileDescription * CPVRFile::GetFileDescription (void)
{
	return &m_sfdFile;
}

bool CPVRFile::WriteFileData(unsigned __int16 nFilenummer)
{
	HANDLE			hFile;
	DWORD			dwBytesWritten;
  char * pFiledata;

  if (m_nFirmwareType == ID_189)
  {
    pFiledata = (char *) &m_sfdFile_old;
    memcpy (m_sfdFile_old.cClipName, m_sfdFile.cClipName, sizeof (m_sfdFile_old.cClipName));
  }
  else
  {
    pFiledata = (char *) &m_sfdFile;
  }
  
	if (OpenDrive(hFile, m_sDrivename))
	{
		SetToPhysicalSector(hFile, TOC_OFFSET_SECTORS + nFilenummer);
		WriteFile(hFile, pFiledata,sizeof(m_sfdFile),&dwBytesWritten,NULL);
  	CloseHandle(hFile);
	}

	return (dwBytesWritten == sizeof (m_sfdFile));
	

}

#endif

bool CPVRFile::SearchForStart(__int64 i64StartPTS, CIBits* pcbStream, unsigned int uiStreamID
		, __int16 iFatEntry, __int16 iClusterCount)
{
	HANDLE hFile;
	long lClusters[MAX_FATPOINTERS];
	int i, j, nTargetCluster (-1), nTargetBuffer (-1), nBubble (50), nSub,
    nSupTCluster (-1), nSupTBuffer (-1);
	unsigned __int32	ui32PESStartCode (0x100 + uiStreamID);
	__int64 i64CurrentPTS, i64LastPTS (0), i64Diff (-1), i64Dist (0), i64PTSPerCluster (0);
  unsigned __int64 ui64FileSize, ui64Temp, ui64LastFilePos (0), ui64FilePos (0);
  bool bFound (false), bFast (false), bFirst (true);
  __int32 nOffset (0), nBytesPerSecond;
	nSub = 27;

  Output ("CPVRFile::SearchForStart\n", false, true);  
	if (!OpenDrive(hFile, m_sDrivename))
  {
    sprintf (m_sOutBuffer, "OpenDrive(hFile %d, Drive %s)... failed\n",hFile, m_sDrivename.c_str ());
    Output (m_sOutBuffer, false, true);  
	  return false;
  }
  sprintf (m_sOutBuffer, "OpenDrive(hFile %d, Drive %s)... success\n",hFile, m_sDrivename.c_str ());
  Output (m_sOutBuffer, false, true);  
	g_pfatFAT->GetClusters (lClusters, iFatEntry);
	// Ersten Cluster finden
  if (iClusterCount > 1)
  {
    ui64FileSize = iClusterCount - 2;
    ui64FileSize *= BYTES_PER_CLUSTER;
    ui64FileSize += (SECTORS_PER_CLUSTER - m_nFirstClusterOffset + m_nLastClusterOffset) * 512;
  }
  else
  {
    ui64FileSize = (SECTORS_PER_CLUSTER - m_nFirstClusterOffset - (SECTORS_PER_CLUSTER - m_nLastClusterOffset)) * 512;
  }
  nBytesPerSecond = (int) (ui64FileSize / (m_dLenInSeconds / 100));
  ui64Temp = nBytesPerSecond;
  ui64Temp *= (unsigned __int64) m_dOffsetTime;
  ui64LastFilePos = ui64Temp;
  j = (int) (ui64Temp / BYTES_PER_CLUSTER);
//  i64PTSPerCluster = (__int64)(m_dLenInSeconds / (double) iClusterCount) * 900;
  nOffset = (int) (ui64Temp % BYTES_PER_CLUSTER);
  nOffset = (nOffset / MBYTE) * MBYTE;
  if (j == 0) nOffset += (m_nFirstClusterOffset * 512);

  sprintf (m_sOutBuffer, "StartPTS 0x%07I64X\n", i64StartPTS);
  Output (m_sOutBuffer, false, true);  
  sprintf (m_sOutBuffer, "Predicted Cluster: %d\n", j);
  Output (m_sOutBuffer, false, true);  
  i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
	m_bOverrun = false;
  if (i64CurrentPTS  < m_i64FirstPTS - 180000 && i64CurrentPTS != -1) m_bOverrun = true;
  pcbStream->ClrEOBS ();
  if (m_bOverrun) i64CurrentPTS += MAX_PTS_33BIT;
  i64LastPTS = i64CurrentPTS;
  i64Diff = (i64CurrentPTS - i64StartPTS) / 90000;
  i64Diff *= nBytesPerSecond;
  sprintf (m_sOutBuffer, "Prediction fault: %I64d\n", i64Diff);
  Output (m_sOutBuffer, false, true);  
  while (i64Diff > MBYTE || i64Diff < -MBYTE)
  {
    if (i64Diff > nOffset)
    {
      j--;
      nOffset = (int) (BYTES_PER_CLUSTER - (i64Diff - nOffset));
    }
    else
    {
      nOffset -= (int) i64Diff;
    }
    if (nOffset > BYTES_PER_CLUSTER)
    {
      j+= nOffset / BYTES_PER_CLUSTER;
      nOffset %= BYTES_PER_CLUSTER;
    }
    nOffset = (nOffset / MBYTE) * MBYTE;
    ui64FilePos = j;
    ui64FilePos *= BYTES_PER_CLUSTER;
    ui64FilePos += nOffset;
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
    if (i64CurrentPTS == -1) break;
    if (i64CurrentPTS  < m_i64FirstPTS - 180000 && i64CurrentPTS != -1)i64CurrentPTS += MAX_PTS_33BIT;
    if ((i64CurrentPTS - i64LastPTS) / 90000 != 0)
    {
      nBytesPerSecond = (int) (i64abs (ui64LastFilePos, ui64FilePos) / (i64abs (i64CurrentPTS, i64LastPTS) / 90000));
    }
    i64LastPTS = i64CurrentPTS;
    ui64LastFilePos = ui64FilePos;
    i64Diff = ((i64CurrentPTS - i64StartPTS) / 90000) + 1;
    i64Diff *= nBytesPerSecond;
    sprintf (m_sOutBuffer, "Prediction fault: %I64d\n", i64Diff);
    Output (m_sOutBuffer, false, true);  
  }
  i64Diff = i64CurrentPTS - i64StartPTS;
  if (i64Diff > 0) 
  {
    while (i64CurrentPTS > i64StartPTS)
    {
      sprintf (m_sOutBuffer, "too far\n", i64Diff);
      Output (m_sOutBuffer, false, true);      
      nOffset -= MBYTE;
      if (nOffset < 0 && j > 0)
      {
        nOffset = 99 * MBYTE;
        j--;
      }
      if (nOffset < 0 && j == 0)
      {
        nOffset = m_nFirstClusterOffset;
        break;
      }
      if (j < 0) 
      {
        j = 0;
        break;
      }
      i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
      if (i64CurrentPTS == -1) break;
      if (i64CurrentPTS  < m_i64FirstPTS - 180000 && i64CurrentPTS != -1)i64CurrentPTS += MAX_PTS_33BIT;
    }
  }

  if (i64CurrentPTS != -1)
  {
	  Output ("Silk and Smooth ..... running\n");
	  i64CurrentPTS = j;
	  i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
    if (j == 0 && nOffset < (m_nFirstClusterOffset * 512))
    {
      i64CurrentPTS += (__int64) ((m_nFirstClusterOffset * 512) * 8);
    }
    else
    {
		  i64CurrentPTS += (__int64) (nOffset * 8);
    }
	  pcbStream->SeekTo (i64CurrentPTS);
 	  CloseHandle(hFile);
	  return true;
  }
// DEFEND
  Output ("Trying old version\n");
 	nTargetCluster = -1;
	i = 0;
  j = 0;
  j = (int) ((double) iClusterCount * ((m_dOffsetTime * 100) / m_dLenInSeconds));
  if (j == 0) nOffset += (m_nFirstClusterOffset * 512);
  sprintf (m_sOutBuffer, "StartPTS 0x%07I64X\n", i64StartPTS);
  Output (m_sOutBuffer, false, true);  
  sprintf (m_sOutBuffer, "Predicted Cluster: %d\n", j);
  Output (m_sOutBuffer, false, true);  
  i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
	m_bOverrun = false;
  if (i64CurrentPTS  < m_i64FirstPTS - 180000 && i64CurrentPTS != -1) i64CurrentPTS += MAX_PTS_33BIT;
  pcbStream->ClrEOBS ();
  if (i64CurrentPTS < i64StartPTS) j++;
  if (i64CurrentPTS > i64StartPTS || i64CurrentPTS == -1) 
  {
    if (m_bOverrun && (MAX_PTS_33BIT - i64CurrentPTS) < i64PTSPerCluster)
      m_bOverrun = false;
    j-=2;
  }
  sprintf (m_sOutBuffer, "Prediction fault: %d\n", i64StartPTS - i64CurrentPTS);
  Output (m_sOutBuffer, false, true);  

  if (j < 0) j = 0;
	for (; j < iClusterCount && !(*m_pbCancel); j++)
	{
    nOffset = 0;
    if (j == 0) nOffset += (m_nFirstClusterOffset * 512);
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
    if (i64CurrentPTS == -1) continue;
		sprintf (m_sOutBuffer, "Searching Cluster %02d\n", j);
		Output (m_sOutBuffer, false, true);
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
    if ((j == 1 && !m_nFirstClusterOffset) || (j > 2) && m_nRipType == RIP_TYPE_VIDEO)
    {
      if (i64Diff == -1) i64Diff = i64CurrentPTS - i64LastPTS;
      i64Dist = i64StartPTS - i64CurrentPTS;
      if (i64Dist > (i64Diff * 2))
        j = j + (int) (i64Dist / (i64Diff * 2));
      bFast = true;
    }
		i64LastPTS = i64CurrentPTS;

		if (i64CurrentPTS > i64StartPTS)
		{
			nTargetCluster = j - 1;
			break;
		}
		if (i64CurrentPTS == i64StartPTS)
		{
      bFound = true;
			nTargetCluster = j;
			break;
		}
		if (i64CurrentPTS == -1)
		{
			j--;
			i += 4;
		}
		else
		{
			i = 0;
			i64LastPTS = i64CurrentPTS;
		}
	}

	// im letzten Cluster
	if (i64CurrentPTS < i64StartPTS && nTargetCluster < 0)
	{
		nTargetCluster = j-1;
	}
	if (nTargetCluster < 0 || m_bStreamWasCut) nTargetCluster = 0;
  if (nTargetCluster >= iClusterCount) nTargetCluster = iClusterCount - 1;
	bFast = false;
	i64LastPTS = 0;
	i = 0;
	m_bOverrun = false;
  i64Diff = -1;
	for (j = 0; j < 100 && !bFound && !(*m_pbCancel); j++)
	{
    nOffset = (j * 0x800 * 512);
    if (j == 0 && nTargetCluster == 0) 
    {
      nOffset += (m_nFirstClusterOffset * 512);
      j = (m_nFirstClusterOffset * 512) / MBYTE;
    }
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[nTargetCluster], nTargetCluster, nOffset, hFile, m_pBuffer);
    if (i64CurrentPTS == -1 && nTargetCluster == iClusterCount - 1 && j == 0)
    {
      pcbStream->ClrEOBS ();
      nTargetCluster--;
      j = 90;
    }
    if (nTargetCluster == iClusterCount - 1 && (j * MBYTE > m_nLastClusterOffset * 512)) // EOB erreicht
    {
      nTargetBuffer = (m_nLastClusterOffset * 512) / MBYTE;
      pcbStream->ClrEOBS ();
      if (i64CurrentPTS != -1)
      {
        i64CurrentPTS = i64StartPTS;
      }
      else
      {
        i64CurrentPTS = i64StartPTS;
      }
      break;
    }
    if (i64CurrentPTS == -1) continue;
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
    if ((j == 1 && nTargetCluster != 0 && !m_nFirstClusterOffset) || (j > 2) && m_nRipType == RIP_TYPE_VIDEO)
    {
      if (i64Diff == -1 && !bFirst) i64Diff = i64CurrentPTS - i64LastPTS;
      i64Dist = i64StartPTS - i64CurrentPTS;
      if (i64Dist > (i64Diff * 1.5) && !bFirst)
        j = j + (int) (i64Dist / (i64Diff * 1.5));
      if (nTargetCluster != 0 || (((j - 1) * 0x100000) > (m_nFirstClusterOffset * 512)))
        bFirst = false;
      bFast = true;
    }
		if (i64CurrentPTS > i64StartPTS && i64CurrentPTS != -1)
		{
			if (!m_bStreamWasCut || (i64abs (i64CurrentPTS, i64StartPTS) < 0x1000000))
			{
				nTargetBuffer = j - 1;
				break;
			}
		}
		if (i64CurrentPTS == i64StartPTS && i64CurrentPTS != -1) 
		{
			nTargetBuffer = j;
      bFound = true;
			break;
		}
		if (i64CurrentPTS == -1)
		{
			j--;
			i += 4;
		}
		else
		{
			i = 0;
			i64LastPTS = i64CurrentPTS;
		}
		i64LastPTS = i64CurrentPTS;
		sprintf (m_sOutBuffer, "Searching Buffer %02d ...\n", j);
		Output (m_sOutBuffer);
	}
	if (!bFound) nTargetBuffer --;
	if (nTargetBuffer < 0) nTargetBuffer = 0;
  if (!bFound) // falls nicht eh schon gefunden
  {
  }
	if (i64CurrentPTS == -1 || i64CurrentPTS < (i64StartPTS >> 1))
	{
		sprintf (m_sOutBuffer, "Shit something is fucked up (i64CurrentPTS = %f) rewinding\n", (double) i64CurrentPTS);
		Output (m_sOutBuffer);
		pcbStream->Rewind ();
		
	}
	else
	{
		Output ("Silk and Smooth ..... running\n");
		i64CurrentPTS = nTargetCluster;
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
    if (nTargetCluster == 0 && (nTargetBuffer * 0x100000) < (m_nFirstClusterOffset * 512))
    {
      i64CurrentPTS += (__int64) ((m_nFirstClusterOffset * 512) * 8);
    }
    else
    {
		  i64CurrentPTS += (__int64) ((nTargetBuffer * 0x100000) * 8);
    }
		pcbStream->SeekTo (i64CurrentPTS);
	}



 	CloseHandle(hFile);
	return true;
}

typedef struct SeekClusters
{
	int nClusterNo, nNextCluster, nId;
	__int64 i64FirstPTS, i64PTSAfter1Buf, i64PTSAfter99Buf;
	bool bFirstCluster, bLastCluster, bUsed;
}
TSeekClusters;

/*
void CPVRFile::LookupClusters()
{
	CStdioFile fOut;
	__int64 i64CurrentPTS;
	unsigned __int64 ui64CurrentPTS;
	int i, nSkip (0), j, nFirst, nLast, nClipNum (1);
	unsigned int nID;
	TSeekClusters sClust[800];
	CString s, s64, sDummy ("00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000");
	CIBits 	* pcbStream = NULL;
	HANDLE hFile;
	DWORD dwRead;
	CPES_Packet * pppPES;
	sFatPointer * pFat;



	if (!OpenDrive(hFile, m_sDrivename))
	return;

	pcbStream=new CIBits(m_sDrivename,1);
	/*
	for ( i = 1; i < 710; i++)
	{

		i64CurrentPTS = -1;
		pcbStream->m_lvecClusters[0] = i;
		GetStreamIDs (*pcbStream, nID, nSkip);
		sClust[i].nClusterNo = i;
		sClust[i].bFirstCluster = false;
		sClust[i].bLastCluster = false;
		sClust[i].bUsed = false;
		sClust[i].i64FirstPTS = -1;
		sClust[i].i64PTSAfter1Buf = -1;
		sClust[i].i64PTSAfter99Buf = -1;
		sClust[i].nId = -1;
		if (nID)
		{
			sClust[i].nId = nID;
		}
		if (*m_pbCancel)
		{
			break;
		}
		sprintf (m_sOutBuffer, "%d : %d", i, nID);
		Output (m_sOutBuffer);

		if (nID)
		{
			GetFirstPTS (pcbStream, nID, i64CurrentPTS);
			sClust[i].i64FirstPTS = i64CurrentPTS;
			sClust[i].nId = nID;
		}
		else
		{
			Sleep (100);
		}

		ui64CurrentPTS = (unsigned __int64) i64CurrentPTS;
		_i64toa (ui64CurrentPTS, s64.GetBuffer (255), 10);
		s64.ReleaseBuffer ();
		//s64 = sDummy.Left (18-s64.GetLength()) + s64;
		if (*m_pbCancel)
		{
			break;
		}

		SetToPhysicalSector(hFile, 0x100 + (i - 1) * 0x32000 + 0x800);
		ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
		j = 0;
		for (; j < 0x100000; j++)
		{
			if (m_pBuffer[j] == 0 && m_pBuffer[j+1] == 0 && m_pBuffer[j+2] == 0x01 && m_pBuffer[j+3] == (char) nID)
				break;
		}

		i64CurrentPTS = 0;
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
		i64CurrentPTS += (__int64) ((j + 0x100000) * 8);
		pcbStream->SeekTo (i64CurrentPTS);
		pppPES = new CPES_Packet(pcbStream, nID);	
		i64CurrentPTS = pppPES->PTS ();
		sClust[i].i64PTSAfter1Buf = i64CurrentPTS;
		delete pppPES;

		SetToPhysicalSector(hFile, 0x100 + (i - 1) * 0x32000 + 99 * 0x800);
		ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
		j = 0;
		for (; j < 0x100000; j++)
		{
			if (m_pBuffer[j] == 0 && m_pBuffer[j+1] == 0 && m_pBuffer[j+2] == 0x01 && m_pBuffer[j+3] == (char) nID)
				break;
		}
		i64CurrentPTS = 0;
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
		i64CurrentPTS += (__int64) ((j + 99 * 0x100000) * 8);
		pcbStream->SeekTo (i64CurrentPTS);
		pppPES = new CPES_Packet(pcbStream, nID);	
		i64CurrentPTS = pppPES->PTS ();
		sClust[i].i64PTSAfter99Buf = i64CurrentPTS;
		sprintf (m_sOutBuffer, "%02d\t%d\t%i64\t%i64\t%i64", i, nID, sClust[i].i64FirstPTS,
			sClust[i].i64PTSAfter1Buf, sClust[i].i64PTSAfter99Buf);
		Output (m_sOutBuffer);
		fOut.WriteString (m_sOutBuffer);
		delete pppPES;
	
		Sleep (100);
	}
	fOut.Open ("Cluster.bin", CFile::modeRead | CFile::typeBinary);
	fOut.Read (sClust, sizeof (sClust));
	fOut.Close ();
	for ( i = 1; i < 710; i++)
	{
		sClust[i].bFirstCluster = true;
		sClust[i].bLastCluster = false;
		sClust[i].bUsed = false;
		sClust[i].nNextCluster = -1;
		sClust[i].nClusterNo = -1;
	}
	nSkip = 710;
	for (i = 1; i < nSkip; i++)
	{
		i64CurrentPTS = sClust[i].i64PTSAfter1Buf - sClust[i].i64FirstPTS; // Durchschnittlicher dingens
		if (sClust[i].i64PTSAfter99Buf == -1) continue;
		if (i64abs (sClust[i].i64FirstPTS + i64CurrentPTS * 99
			, sClust[i].i64PTSAfter99Buf) < 0x100000) continue;
		for (j = 1; j < nSkip; j++)
		{
			if (i == j) continue;
			if (sClust[j].bUsed) continue;
			if (i64abs (sClust[i].i64PTSAfter99Buf + i64CurrentPTS, sClust[j].i64FirstPTS) < i64CurrentPTS *0xA 
				&& sClust[i].nId == sClust[j].nId)
			{
				sClust[i].nNextCluster = j;
				sClust[j].bUsed = true;
				sClust[j].bFirstCluster = false;
				j = nSkip;
			}
		}
	}
	// Matching Audio and video
	for (i = 1; i < nSkip; i++)
	{
		if (sClust[i].bFirstCluster && sClust[i].nId >= 0xe0) // First Video
		{
			for (j = 1; j < nSkip; j++)
			{
				if (j == i) continue;
				if (sClust[j].nId >= 0xe0) continue;
				if (i64abs (sClust[i].i64FirstPTS, sClust[j].i64FirstPTS) < 90000)
				{
					sClust[i].nClusterNo = j;
					j = nSkip;
				}
			}
		}
	}


	fOut.Open ("bFound.txt", CFile::modeCreate | CFile::modeWrite | CFile::typeText);
	
	for (i = 0; i < nSkip; i++)
	{
		if (sClust[i].nId < 0xe0) continue;
		if (sClust[i].bFirstCluster && sClust[i].nClusterNo != -1)
		{
			nFirst = i;
			m_sfdFile.iVideoClusterCount = 1;
			m_sfdFile.iVideoFatEntry = i;
			s.Format ("Streamtyp : %02X  Clusters : %03d  ", sClust[i].nId,i);
			fOut.WriteString (s);
			j = sClust[i].nNextCluster;
			pFat = g_pfatFAT->GetFatPointer (i);
			pFat->uiNextFatPointer = sClust[i].nNextCluster;
			while (j != -1)
			{
				m_sfdFile.iVideoClusterCount++;
				pFat = g_pfatFAT->GetFatPointer (j);
				s.Format ("%03d  ",j);
				fOut.WriteString (s);
				if (sClust[j].nNextCluster == -1) nLast = j;
				j = sClust[j].nNextCluster;
				pFat->uiNextFatPointer = j;
			}
			fOut.WriteString ("\n");
			fOut.WriteString ("Matching Audio:  ");

			j = sClust[i].nClusterNo;
			m_sfdFile.iAudioFatEntry = j;
			m_sfdFile.iAudioClusterCount = 0;
			while (j != -1)
			{
				m_sfdFile.iAudioClusterCount++;
				pFat = g_pfatFAT->GetFatPointer (j);
				s.Format ("%03d  ",j);
				fOut.WriteString (s);
				j = sClust[j].nNextCluster;
				pFat->uiNextFatPointer = j;
			}
			fOut.WriteString ("\n");
			s.Format ("Clip %02d", nClipNum);
			strcpy (m_sfdFile.cClipName, s);
			strcpy (m_sfdFile.cStationName, s);
			int nLen = (sClust[nLast].i64FirstPTS - sClust[nFirst].i64FirstPTS) / 90000;
			m_sfdFile.iLengthHours = nLen / 3600;
			nLen %= 3600;
			m_sfdFile.iLengthMinutes = nLen / 60;
			m_sfdFile.iLengthSeconds = nLen % 60;
			g_pfatFAT->WriteFatData ();
			WriteFileData (nClipNum++);	
			if (nClipNum > 90) break;


		}
	}
	fOut.Close ();
	fOut.Open ("Cluster.bin", CFile::modeCreate | CFile::modeWrite | CFile::typeBinary);
	fOut.Write (sClust, sizeof (sClust));
	fOut.Close ();
//	delete pcbStream;
}
*/
void CPVRFile::RipAudioStreamRaw (__int64 i64StartPTS,__int64 i64EndPTS, CIBits* pcbStream, unsigned int uiStreamID)
{
  int nOffset;
	HANDLE hFile;
	CFile fOut;
  bool bEnd (false);
	int i(0), j, nLen, nNumberOfClusters, nFirstCluster, nBuffersToSkip (-1), nPacketSize (0);
	int nClustersToRead (0), nBuffersToRead (0);
  OFSTRUCT of;
  CString sFilename (Filename().c_str()), sExt, s, sNum;
	long lClusters[MAX_FATPOINTERS];
	DWORD dwRead;
	unsigned __int32	ui32PESStartCode (0x100 + uiStreamID);
	__int64 i64CurrentPTS, i64LastPTS (0);
	CPES_Packet * pppPES;

	m_nRipType = RIP_TYPE_AUDIO;
	m_bOverrun = false;
	if (!OpenDrive(hFile, m_sDrivename))
	return;
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iAudioFatEntry);

	// Get Filename
  sExt = "." + m_pInfoAudio->m_sPrefixRaw;
  s = sFilename + sExt;
  while (OpenFile (s, &of, OF_EXIST) != HFILE_ERROR && sFilename != "PRV")
  {
    sNum.Format ("_%03d", i++);
    s = sFilename + sNum + sExt;
  }
	if (!fOut.Open (s, CFile::modeCreate | CFile::typeBinary | CFile::modeWrite))
	{
    CloseHandle (hFile);
		return;
	}
	for (nNumberOfClusters = 0; lClusters[nNumberOfClusters] != INVALID_POINTER; nNumberOfClusters++)
	{
	}
	m_pInfoAudio->m_nEndPos = nNumberOfClusters * 100;
	m_pInfoAudio->m_nStartPos = 0;
	m_pInfoAudio->m_nPosition = 0;
	m_pInfoAudio->m_nFileSize = 0;
	Output ("Seeking End of RawAudio ....");
	// die Rohcluster (nicht letzter)
	nFirstCluster = -1;
	nClustersToRead = -1;
	i = 0;
	m_bOverrun = false;
	m_bBeginOverrun = false;
	m_bEndOverrun = false;
	i64LastPTS = 0;
	for (j = 0; j < nNumberOfClusters && !(*m_pbCancel); j++)
	{
    nOffset = 0;
    if (j == 0) nOffset += (m_nFirstClusterOffset * 512);
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
    if (i64CurrentPTS == -1) break;
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
		i64LastPTS = i64CurrentPTS;
		if (i64CurrentPTS > i64StartPTS && nFirstCluster == -1) 
		{
			nFirstCluster = j - 1;
			if (m_bOverrun) m_bBeginOverrun = true;
			else m_bBeginOverrun = false;
		}
		if (i64CurrentPTS == i64StartPTS && nFirstCluster > j) 
		{
			nFirstCluster = j;
			if (m_bOverrun) m_bBeginOverrun = true;
			else m_bBeginOverrun = false;
		}
		// Letzter Cluster und noch nichts gefunden
		if (i64CurrentPTS < i64StartPTS && nFirstCluster == -1 && j == nNumberOfClusters -1)
		{
			nFirstCluster = j;
			if (m_bOverrun) m_bBeginOverrun = true;
			else m_bBeginOverrun = false;
		}
		if (i64CurrentPTS > i64EndPTS) 
		{
			nClustersToRead = j - 1;
			if (m_bOverrun) m_bEndOverrun = true;
			else m_bEndOverrun = false;
			break;
		}
		if (i64CurrentPTS == i64EndPTS) 
		{
			nClustersToRead = j;
			if (m_bOverrun) m_bEndOverrun = true;
			else m_bEndOverrun = false;
			break;
		}
	}
	if (m_bOverrun) m_bEndOverrun = true;
	if (nFirstCluster == -1) nFirstCluster = nNumberOfClusters - 1;
	if (nFirstCluster < 0) nFirstCluster = 0;
	if (nClustersToRead == -1) nClustersToRead = nNumberOfClusters - 1;
	if (nClustersToRead < 0) nClustersToRead = 0;
	if (nFirstCluster > nClustersToRead) nClustersToRead = nFirstCluster;
		
	// Zurck zum Anfang des ersten Clusters (suchen der buffers)
	i = 0;
	m_bOverrun = false;
	for (j = 0; j < 100 && !(*m_pbCancel); j++)
	{
    nOffset = j * MBYTE;
    if (j == 0 && nFirstCluster == 0)
    {
      j = (m_nFirstClusterOffset * 512) / MBYTE;
      nOffset += (m_nFirstClusterOffset * 512);
    }
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[nFirstCluster], nFirstCluster, nOffset, hFile, m_pBuffer);
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
		if (((j * MBYTE > m_nLastClusterOffset * 512) && nFirstCluster == nNumberOfClusters - 1)
      || i64CurrentPTS >= i64StartPTS) 
			break;
		if (i64CurrentPTS == -1)
		{
			continue;
		}
		else
		{
			i64LastPTS = i64CurrentPTS;
			i = 0;
		}


	}
	if (i64CurrentPTS == i64StartPTS) nBuffersToSkip = j;
	if (i64CurrentPTS > i64StartPTS) nBuffersToSkip = j-1;
	if (nBuffersToSkip < 0) nBuffersToSkip = 0;
	i = 0;
	i64LastPTS = 0;
	// Zurck zum Anfang des letzten Clusters
	for (j = 0; j < 100 && !(*m_pbCancel); j++)
	{
    nOffset = j * 0x800 * 512;
    if (j == 0 && nClustersToRead == 0)
    {
      j = (m_nFirstClusterOffset * 512) / MBYTE;
      nOffset += (m_nFirstClusterOffset * 512);
    }
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[nClustersToRead], nClustersToRead, nOffset, hFile, m_pBuffer);
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}

    if (((j * MBYTE > m_nLastClusterOffset * 512) && nClustersToRead == nNumberOfClusters - 1)
      || i64CurrentPTS >= i64EndPTS || (i64CurrentPTS < i64LastPTS && i64CurrentPTS != -1)) 
			break;
		if (i64CurrentPTS == -1)
		{
			j--;
			i += 4;
			continue;
		}
		else
		{
			i64LastPTS = i64CurrentPTS;
			i = 0;
		}

	}
	nBuffersToRead = j;
	if (i64CurrentPTS == i64EndPTS) nBuffersToRead = j;
	
	m_pInfoAudio->m_nEndPos = (nClustersToRead - nFirstCluster) * 100 + (nBuffersToRead - nBuffersToSkip);
	m_pInfoAudio->m_bRunning = true;
	m_pInfoAudio->m_bFirstUpdate = true;

	Output ("Ripping Raw Audio ....");
	
	// Zum ersten Cluster
	i = nFirstCluster;
	SetToPhysicalSector(hFile, 0x100 + (lClusters[nFirstCluster] - 1) * 0x32000 + nBuffersToSkip * 0x800);
	// erster Puffer
	ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
	i = 0;
  j = 0;
	m_bOverrun = false;
	while (!(*m_pbCancel))
	{
		for (; i < 0x100000; i++)
		{
			if (m_pBuffer[i] == 0 && m_pBuffer[i+1] == 0 && m_pBuffer[i+2] == 0x01 && m_pBuffer[i+3] == (BYTE) uiStreamID)
				break;
		}
		if (i == 0x100000) break;
		i64CurrentPTS = (__int64) (nFirstCluster);
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
		i64CurrentPTS += (__int64) ((i + (nBuffersToSkip * 0x100000)) * 8);
		pcbStream->SeekTo (i64CurrentPTS);
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
		i64CurrentPTS = pppPES->PTS ();
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
		if (i64abs (i64CurrentPTS, i64StartPTS) <= 10000 || i64CurrentPTS > i64StartPTS) 
		{
			i64StartPTS = i64CurrentPTS;
			fOut.Write (&m_pBuffer[i], 0x100000-i);
			delete pppPES;
			break;
		}
		if (!(i64CurrentPTS == -1))
			nPacketSize = pppPES->PES_Packet_Length() + 6;
		nLen = i;
		i += 4;
		delete pppPES;
	}
	// erster Cluster Puffer schreiben
	i = nFirstCluster;
	SetToPhysicalSector(hFile, 0x100 + (lClusters[i] - 1) * 0x32000 + (nBuffersToSkip + 1) * 0x800);
	for (i = nBuffersToSkip + 1; i < 100 && !(*m_pbCancel) ; i++)
	{
		if (nClustersToRead - nFirstCluster == 0  && i >= nBuffersToRead) break;
		ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
		fOut.Write (m_pBuffer, 0x100000);
		m_pInfoAudio->m_nPosition += 1;
		m_pInfoAudio->m_nFileSize += 1;

	}		


	for (i = nFirstCluster+1; i < nClustersToRead && !(*m_pbCancel); i++)
	{
		SetToPhysicalSector(hFile, 0x100 + (lClusters[i] - 1) * 0x32000);
		for (nLen = 0; nLen < 100; nLen++)
		{

			if (*m_pbCancel)
			{
				CloseHandle(hFile);
				fOut.Close ();
				Output ("cancelled");
        CloseHandle (hFile);
				return;
			}

			m_pInfoAudio->m_nPosition = (i - nFirstCluster - 1) * 100 + nLen;
			m_pInfoAudio->m_nFileSize = (i - nFirstCluster - 1) * 100 + nLen;
			ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
			fOut.Write (m_pBuffer, 0x100000);
		}
	}
	// Letzter Cluster (Endeerkennung)
	if (nClustersToRead - nFirstCluster != 0 && !(*m_pbCancel)) 
	{
    if (nClustersToRead == nNumberOfClusters-1)
    {
      if (nBuffersToRead > (m_sfdFile.iAudioStopFirstCluster * 512) / MBYTE + 1)
        nBuffersToRead = (m_sfdFile.iAudioStopFirstCluster * 512) / MBYTE + 1;
    }
    if (nBuffersToRead > 100) nBuffersToRead = 100;
		SetToPhysicalSector(hFile, 0x100 + (lClusters[nClustersToRead] - 1) * 0x32000);
		for (i = 0; i < nBuffersToRead; i++)
		{
			
			ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
			fOut.Write (m_pBuffer, 0x100000);
			m_pInfoAudio->m_nPosition += 1;
			m_pInfoAudio->m_nFileSize += 1;

		}		
	}
/*	// letzter Puffer
  if (!(*m_pbCancel))
	  ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
	i = 0;
	nLen = 0;
	while (!(*m_pbCancel))
	{
		for (; i < 0x100000; i++)
		{
			if (m_pBuffer[i] == 0 && m_pBuffer[i+1] == 0 && m_pBuffer[i+2] == 0x01 && m_pBuffer[i+3] == (BYTE) uiStreamID)
				break;
		}
		if (i >= 0x100000) break;
		i64CurrentPTS = (__int64) (nClustersToRead);
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
		i64CurrentPTS += (__int64) ((i + (nBuffersToRead  * 0x100000)) * 8);
    if (i64CurrentPTS > (__int64) m_i64AudioMaxPos) break;
		pcbStream->SeekTo (i64CurrentPTS);
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
		i64CurrentPTS = pppPES->PTS ();
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
		if ((i64CurrentPTS > i64EndPTS) || (i64LastPTS > i64CurrentPTS)) 
		{
      break;
		}
		if (i64CurrentPTS == -1)
		{
			j--;
			i+=4;
		}
		else
		{
			i64LastPTS = i64CurrentPTS;
			nPacketSize = pppPES->PES_Packet_Length() + 6;
			nLen = i;
			i += 4;
		}
		delete pppPES;
	}
  if (!(*m_pbCancel))
  {
  	fOut.Write (m_pBuffer, nLen + nPacketSize);
  }
*/
	m_pInfoAudio->m_nPosition += 1;
	m_pInfoAudio->m_nFileSize += 1;
	
	CloseHandle(hFile);
	fOut.Close ();
	Output ("done ....");

}

void CPVRFile::RipVideoStreamRaw (__int64 i64StartPTS,__int64 i64EndPTS, CIBits* pcbStream, unsigned int uiStreamID)
{

	HANDLE hFile;
	CFile fOut;
	int i(0), j(0), nLen(0), nNumberOfClusters(0), nFirstCluster(0), nBuffersToSkip(0), nPacketSize(0);
	int nClustersToRead (0), nBuffersToRead (0);
  OFSTRUCT of;
  CString sFilename (Filename().c_str()), sExt, s, sNum;
  int nOffset;
	long lClusters[MAX_FATPOINTERS];
	DWORD dwRead;
	unsigned __int32	ui32PESStartCode (0x100 + uiStreamID);
	__int64 i64CurrentPTS, i64LastPTS (0);
	CPES_Packet * pppPES;

	m_bOverrun = false;
	if (!OpenDrive(hFile, m_sDrivename))
	return;
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iVideoFatEntry);

	// Get Filename
  sExt = "." + m_pInfoVideo->m_sPrefixRaw;
  s = sFilename + sExt;
  while (OpenFile (s, &of, OF_EXIST) != HFILE_ERROR)
  {
    sNum.Format ("_%03d", i++);
    s = sFilename + sNum + sExt;
  }
	if (!fOut.Open (s, CFile::modeCreate | CFile::typeBinary | CFile::modeWrite))
	{
		return;
	}
	for (nNumberOfClusters = 0; lClusters[nNumberOfClusters] != INVALID_POINTER; nNumberOfClusters++)
	{
	}
	m_pInfoVideo->m_nEndPos = nNumberOfClusters * 100;
	m_pInfoVideo->m_nStartPos = 0;
	m_pInfoVideo->m_nPosition = 0;
	m_pInfoVideo->m_nFileSize = 0;
	Output ("Seeking End of RawVideo ....");
	// die Rohcluster (nicht letzter)
	nFirstCluster = -1;
	nClustersToRead = -1;
	i = 0;
	m_bOverrun = false;
	m_bBeginOverrun = false;
	m_bEndOverrun = false;
	i64LastPTS = 0;
	for (j = 0; j < nNumberOfClusters && !(*m_pbCancel); j++)
	{
    nOffset = 0;
    if (j == 0) nOffset += (m_nFirstClusterOffset * 512);
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[j], j, nOffset, hFile, m_pBuffer);
    if (i64CurrentPTS == -1) break;
    // berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
		i64LastPTS = i64CurrentPTS;
		if (i64CurrentPTS > i64StartPTS && nFirstCluster == -1) 
		{
			nFirstCluster = j - 1;
			if (m_bOverrun) m_bBeginOverrun = true;
			else m_bBeginOverrun = false;
		}
		if (i64CurrentPTS == i64StartPTS && nFirstCluster > j) 
		{
			nFirstCluster = j;
			if (m_bOverrun) m_bBeginOverrun = true;
			else m_bBeginOverrun = false;
		}
		// Letzter Cluster und noch nichts gefunden
		if (i64CurrentPTS < i64StartPTS && nFirstCluster == -1 && j == nNumberOfClusters -1)
		{
			nFirstCluster = j;
			if (m_bOverrun) m_bBeginOverrun = true;
			else m_bBeginOverrun = false;
		}
		if (i64CurrentPTS > i64EndPTS) 
		{
			nClustersToRead = j - 1;
			if (m_bOverrun) m_bEndOverrun = true;
			else m_bEndOverrun = false;
			break;
		}
		if (i64CurrentPTS == i64EndPTS) 
		{
			nClustersToRead = j;
			if (m_bOverrun) m_bEndOverrun = true;
			else m_bEndOverrun = false;
			break;
		}
	}
	if (m_bOverrun) m_bEndOverrun = true;
	if (nFirstCluster == -1) nFirstCluster = nNumberOfClusters - 1;
	if (nFirstCluster < 0) nFirstCluster = 0;
	if (nClustersToRead == -1) nClustersToRead = nNumberOfClusters - 1;
	if (nClustersToRead < 0) nClustersToRead = 0;
	if (nFirstCluster > nClustersToRead) nClustersToRead = nFirstCluster;
		
	// Zurck zum Anfang des ersten Clusters (suchen der buffers)
	i = 0;
	m_bOverrun = false;
	for (j = 0; j < 100 && !(*m_pbCancel); j++)
	{
    nOffset = j * MBYTE;
    if (j == 0 && nFirstCluster == 0)
    {
      j = (m_nFirstClusterOffset * 512) / MBYTE;
      nOffset += (m_nFirstClusterOffset * 512);
    }
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[nFirstCluster], nFirstCluster, nOffset, hFile, m_pBuffer);
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}

    if (i == 0x100000 || i64CurrentPTS >= i64StartPTS) 
			break;
		if (i64CurrentPTS == -1)
		{
			j--;
			i = i + 4;
			continue;
		}
		else
		{
			i64LastPTS = i64CurrentPTS;
			i = 0;
		}


	}
	if (i64CurrentPTS == i64StartPTS) nBuffersToSkip = j;
	if (i64CurrentPTS > i64StartPTS) nBuffersToSkip = j-1;
	if (nBuffersToSkip < 0) nBuffersToSkip = 0;
	i = 0;
	i64LastPTS = 0;
	// Zurck zum Anfang des letzten Clusters
	for (j = 0; j < 100 && !(*m_pbCancel); j++)
	{
    nOffset = j * 0x800 * 512;
    if (j == 0 && nClustersToRead == 0)
    {
      j = (m_nFirstClusterOffset * 512) / MBYTE;
      nOffset += (m_nFirstClusterOffset * 512);
    }
    i64CurrentPTS = GetNextPTS (uiStreamID, pcbStream, lClusters[nClustersToRead], nClustersToRead, nOffset, hFile, m_pBuffer);
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}

    if (((j * MBYTE > m_nLastClusterOffset * 512) && nClustersToRead == nNumberOfClusters - 1)
      || i64CurrentPTS >= i64EndPTS || (i64CurrentPTS < i64LastPTS && i64CurrentPTS != -1)) 
			break;
		if (i64CurrentPTS == -1)
		{
			j--;
			i += 4;
			continue;
		}
		else
		{
			i64LastPTS = i64CurrentPTS;
			i = 0;
		}

	}
	nBuffersToRead = j-1;
	if (i64CurrentPTS == i64EndPTS) nBuffersToRead = j;

	m_pInfoVideo->m_nEndPos = (nClustersToRead - nFirstCluster) * 100 + (nBuffersToRead - nBuffersToSkip);
	m_pInfoVideo->m_bRunning = true;
	m_pInfoVideo->m_bFirstUpdate = true;
	m_nRipType = RIP_TYPE_VIDEO;

	Output ("Ripping Raw Video ....");
	
	// Zum ersten Cluster
	i = nFirstCluster;
	SetToPhysicalSector(hFile, 0x100 + (lClusters[nFirstCluster] - 1) * 0x32000 + nBuffersToSkip * 0x800);
	// erster Puffer
	if (!(*m_pbCancel)) ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
	i = 0;
	m_bOverrun = false;
	while (!(*m_pbCancel))
	{
		for (; i < 0x100000 ; i++)
		{
			if (m_pBuffer[i] == 0 && m_pBuffer[i+1] == 0 && m_pBuffer[i+2] == 0x01 && m_pBuffer[i+3] == (BYTE) uiStreamID)
				break;
		}
		if (i == 0x100000) break;
		i64CurrentPTS = (__int64) (nFirstCluster);
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
		i64CurrentPTS += (__int64) ((i + (nBuffersToSkip * 0x100000)) * 8);
		pcbStream->SeekTo (i64CurrentPTS);
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
		i64CurrentPTS = pppPES->PTS ();
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}

    if (i64abs (i64CurrentPTS, i64StartPTS) <= 15000 || i64CurrentPTS > i64StartPTS) 
		{
			i64StartPTS = i64CurrentPTS;
			fOut.Write (&m_pBuffer[i], 0x100000-i);
			delete pppPES;
			break;
		}
		if (!(i64CurrentPTS == -1))
			nPacketSize = pppPES->PES_Packet_Length() + 6;
		nLen = i;
		i += 4;
		delete pppPES;
	}
	// erster Cluster Puffer schreiben
	i = nFirstCluster;
	SetToPhysicalSector(hFile, 0x100 + (lClusters[i] - 1) * 0x32000 + (nBuffersToSkip + 1) * 0x800);
	for (i = nBuffersToSkip + 1; i < 100 && !(*m_pbCancel); i++)
	{
		if (nClustersToRead - nFirstCluster == 0  && i >= nBuffersToRead) break;
		ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
		fOut.Write (m_pBuffer, 0x100000);
		m_pInfoVideo->m_nPosition += 1;
		m_pInfoVideo->m_nFileSize += 1;

	}

	for (i = nFirstCluster+1; i < nClustersToRead && !(*m_pbCancel); i++)
	{
		SetToPhysicalSector(hFile, 0x100 + (lClusters[i] - 1) * 0x32000);
		for (nLen = 0; nLen < 100; nLen++)
		{

			if (*m_pbCancel)
			{
				CloseHandle(hFile);
				fOut.Close ();
				Output ("cancelled");
				return;
			}

			m_pInfoVideo->m_nPosition = (i - nFirstCluster) * 100 + nLen;
			m_pInfoVideo->m_nFileSize = (i - nFirstCluster) * 100 + nLen;
			ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
			fOut.Write (m_pBuffer, 0x100000);
		}
	}
	// Letzter Cluster (Endeerkennung)
	if (nClustersToRead - nFirstCluster != 0 && !(*m_pbCancel)) 
	{
		SetToPhysicalSector(hFile, 0x100 + (lClusters[nClustersToRead] - 1) * 0x32000);
		for (i = 0; i < nBuffersToRead; i++)
		{
			
			ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
			fOut.Write (m_pBuffer, 0x100000);
			m_pInfoVideo->m_nPosition += 1;
			m_pInfoVideo->m_nFileSize += 1;

		}		
	}
	// letzter Puffer
	if (!(*m_pbCancel)) ReadFile (hFile, m_pBuffer, 0x100000, &dwRead, NULL);
	i = 0;
	nLen = 0;
	while (!(*m_pbCancel))
	{
		for (; i < 0x100000; i++)
		{
			if (m_pBuffer[i] == 0 && m_pBuffer[i+1] == 0 && m_pBuffer[i+2] == 0x01 && m_pBuffer[i+3] == (BYTE) uiStreamID)
				break;
		}
		if (i >= 0x100000) break;
		i64CurrentPTS = (__int64) (nClustersToRead);
		i64CurrentPTS *= (__int64) BITS_PER_CLUSTER;
		i64CurrentPTS += (__int64) ((i + (nBuffersToRead  * 0x100000)) * 8);
    if (i64CurrentPTS > (__int64) m_i64VideoMaxPos) break;
		pcbStream->SeekTo (i64CurrentPTS);
		pppPES = new CPES_Packet(pcbStream, uiStreamID);
		i64CurrentPTS = pppPES->PTS ();
		// berlauf abfangen
		if (i64CurrentPTS < (m_i64FirstPTS - 180000) && i64CurrentPTS != -1)
		{
			i64CurrentPTS += MAX_PTS_33BIT;
			m_bOverrun = true;
		}
		if ((i64CurrentPTS > i64EndPTS) || (i64LastPTS > i64CurrentPTS)) 
		{
			break;
		}
		if (i64CurrentPTS == -1)
		{
			j--;
			i+=4;
		}
		else
		{
			i64LastPTS = i64CurrentPTS;
			nPacketSize = pppPES->PES_Packet_Length() + 6;
			nLen = i;
			i += 4;
		}
		delete pppPES;
	}
  if (!(*m_pbCancel))
  {
	  fOut.Write (m_pBuffer, nLen + nPacketSize);
  }
	m_pInfoVideo->m_nPosition += 1;
	m_pInfoVideo->m_nFileSize += 1;
	
	CloseHandle(hFile);
	fOut.Close ();
	Output ("done ....");

}


void CPVRFile::SetFilename(const char *pFilename, bool bTake)
{
	if (pFilename)
	{
		strcpy (m_cFilename, pFilename);
    if (bTake)
    {
		  strncpy (m_sfdFile.cClipName, pFilename, 39);
		  m_sfdFile.cClipName[40] = 0;
		  WriteFileData (m_nFileNo);
    }
	}
	else
	{
		m_cFilename[0] = 0;
	}
}


void CPVRFile::dump_file()
{
	CString s;
	CStdioFile fOut;
	int i;
	s.Format ("%d_", m_sfdFile.iRecordNo+1);
	s += m_sfdFile.cClipName;
	s += ".log";
	if (!fOut.Open (s, CFile::modeCreate | CFile::modeWrite | CFile::typeText)) return;
	s.Format ("Record: %02d\n", m_sfdFile.iRecordDay);
	fOut.WriteString (s);
	fOut.WriteString (m_sfdFile.cStationName);
	fOut.WriteString ("\n");
	fOut.WriteString (m_sfdFile.cClipName);
	fOut.WriteString ("\n");
	s.Format ("Record: %02d\n", m_sfdFile.cUnknown1);
	fOut.WriteString (s);
	fOut.WriteString ("Unknown 2 - 1 ");
	for (i = 0; i < 5; i++)
	{
		s.Format ("%02x ", (BYTE) m_sfdFile.cUnknown2_1[i]);
		fOut.WriteString (s);
	}
	fOut.WriteString ("\n");
	fOut.WriteString ("Video Stop first");
  s.Format ("%04X ", m_sfdFile.iVideoStopFirstCluster);
	fOut.WriteString ("\n");
	fOut.WriteString ("Audio Stop first");
  s.Format ("%04X ", m_sfdFile.iAudioStopFirstCluster);
	fOut.WriteString ("\n");
	fOut.WriteString ("Remain 2 - 1 ");
	for (i = 0; i < 0x180; i++)
	{
		s.Format ("%02x ", (BYTE) m_sfdFile.cRemaining[i]);
		fOut.WriteString (s);
	}
	fOut.WriteString ("\n");
	fOut.Close ();
	
}


void CPVRFile::BackupClusters()
{
	CString sFilename, sTemp;
	HANDLE hInFile;
	CFile fOut;
	int i, j;
	long lClusters[MAX_FATPOINTERS];
	DWORD dwRead, tick=0,tack=0;

	// erst mal das Fileinfo wegschreiben
	sFilename = Filename ().c_str ();
	sTemp = sFilename + ".sFileDescription";
	if (!fOut.Open (sTemp, CFile::modeCreate | CFile::modeWrite | CFile::typeBinary))
	{
		Output ("Backup File could not be opened");
		return;
	}
	fOut.Write (&m_sfdFile, sizeof (m_sfdFile));
	fOut.Close ();

	m_pInfoVideo->m_i64StartPos = 0;
	m_pInfoVideo->m_i64EndPos = 0;
	// Drive ffnen
	if (!OpenDrive(hInFile, m_sDrivename))
	return;
	// erst die Audiocluster
	m_nRipType = RIP_TYPE_AUDIO;
	m_pInfoAudio->m_nEndPos = (m_sfdFile.iAudioClusterCount - 1) * 100 
    - (m_sfdFile.iAudioStartFirstCluster * 512 / MBYTE) + (m_sfdFile.iAudioStopFirstCluster * 512 / MBYTE);
	m_pInfoAudio->m_nStartPos = 0;
	m_pInfoAudio->m_nPosition = 0;
	m_pInfoAudio->m_nFileSize = 0;
	m_pInfoAudio->m_bRunning = true;
	m_pInfoAudio->m_bFirstUpdate = true;

	sTemp.Format ("Backing up Audio Clusters (%d MB)", m_pInfoAudio->m_nEndPos);
	Output (sTemp.GetBuffer (255));
	sTemp.ReleaseBuffer ();
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iAudioFatEntry);
	// Zeit beginn
	tick = GetTickCount();
	sTemp.Format ("%s.mpp", sFilename);
	if (!fOut.Open (sTemp, CFile::modeCreate | CFile::modeWrite | CFile::typeBinary))
	{
		Output ("Audio Backup File could not be opened");
		tack = GetTickCount() - tick;
		m_pInfoAudio->m_nRipTime = (int)tack / 1000;
		return;
	}
  // First Buffer Audio
	SetToPhysicalSector(hInFile, 0x100 + (lClusters[0] - 1) * 0x32000 + m_sfdFile.iAudioStartFirstCluster);
	ReadFile (hInFile, m_pBuffer, 0x100000 - ((m_sfdFile.iAudioStartFirstCluster * 512) % 0x100000), &dwRead, NULL);
	fOut.Write (m_pBuffer, 0x100000 - ((m_sfdFile.iAudioStartFirstCluster * 512) % 0x100000));
	m_pInfoAudio->m_nPosition++;
	m_pInfoAudio->m_nFileSize++;
  // First Cluster Audio
  j = ((m_sfdFile.iAudioStartFirstCluster * 512) / 0x100000)+1;
  if (m_sfdFile.iAudioClusterCount > 1)
  {
    for (; j < 100; j++)
	  {
		  ReadFile (hInFile, m_pBuffer, 0x100000, &dwRead, NULL);
		  fOut.Write (m_pBuffer, 0x100000);
		  m_pInfoAudio->m_nPosition++;
		  m_pInfoAudio->m_nFileSize++;
	  }
  }
  // Rest Clusters Audio
	for (i = 1; i < m_sfdFile.iAudioClusterCount-1; i++)
	{
		SetToPhysicalSector(hInFile, 0x100 + (lClusters[i] - 1) * 0x32000);
		for (j = 0; j < 100; j++)
		{
			ReadFile (hInFile, m_pBuffer, 0x100000, &dwRead, NULL);
			fOut.Write (m_pBuffer, 0x100000);
			m_pInfoAudio->m_nPosition++;
			m_pInfoAudio->m_nFileSize++;
		}
	}
  // Last Cluster Audio
	SetToPhysicalSector(hInFile, 0x100 + (lClusters[m_sfdFile.iAudioClusterCount-1] - 1) * 0x32000);
	for (j = 0; j < (m_sfdFile.iAudioStopFirstCluster * 512) / 0x100000 ; j++)
	{
		ReadFile (hInFile, m_pBuffer, 0x100000, &dwRead, NULL);
		fOut.Write (m_pBuffer, 0x100000);
		m_pInfoAudio->m_nPosition++;
		m_pInfoAudio->m_nFileSize++;
	}
	ReadFile (hInFile, m_pBuffer, (m_sfdFile.iAudioStopFirstCluster * 512) % 0x100000, &dwRead, NULL);
	fOut.Write (m_pBuffer, (m_sfdFile.iAudioStopFirstCluster * 512) % 0x100000);

	fOut.Close ();
	tack = GetTickCount() - tick;
	m_pInfoAudio->m_nRipTime = (int)tack / 1000;


	// jetzt die Videocluster
	m_nRipType = RIP_TYPE_VIDEO;
	m_pInfoVideo->m_nEndPos = (m_sfdFile.iVideoClusterCount - 1) * 100 
    - (m_sfdFile.iVideoStartFirstCluster * 512 / MBYTE) + (m_sfdFile.iVideoStopFirstCluster * 512 / MBYTE);
	m_pInfoVideo->m_nStartPos = 0;
	m_pInfoVideo->m_nPosition = 0;
	m_pInfoVideo->m_nFileSize = 0;
	m_pInfoVideo->m_bRunning = true;
	m_pInfoVideo->m_bFirstUpdate = true;
	sTemp.Format ("Backing up Video Clusters (%d MB)", m_pInfoVideo->m_nEndPos);
	Output (sTemp.GetBuffer (255));
	sTemp.ReleaseBuffer ();
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iVideoFatEntry);
	tick = GetTickCount();
	sTemp.Format ("%s.pva", sFilename);
	if (!fOut.Open (sTemp, CFile::modeCreate | CFile::modeWrite | CFile::typeBinary))
	{
		Output ("Video Backup File could not be opened");
		tack = GetTickCount() - tick;
		m_pInfoVideo->m_nRipTime = (int)tack / 1000;
		return;
	}
  // First Buffer Video
	SetToPhysicalSector(hInFile, 0x100 + (lClusters[0] - 1) * 0x32000 + m_sfdFile.iVideoStartFirstCluster);
	ReadFile (hInFile, m_pBuffer, 0x100000 - ((m_sfdFile.iVideoStartFirstCluster * 512) % 0x100000), &dwRead, NULL);
	fOut.Write (m_pBuffer, 0x100000 - ((m_sfdFile.iVideoStartFirstCluster * 512) % 0x100000));
	m_pInfoVideo->m_nPosition++;
	m_pInfoVideo->m_nFileSize++;
  // First Cluster Video
  j = ((m_sfdFile.iVideoStartFirstCluster * 512) / 0x100000)+1;
  if (m_sfdFile.iVideoClusterCount > 1)
  {
	  for (; j < 100; j++)
	  {
		  ReadFile (hInFile, m_pBuffer, 0x100000, &dwRead, NULL);
		  fOut.Write (m_pBuffer, 0x100000);
		  m_pInfoVideo->m_nPosition++;
		  m_pInfoVideo->m_nFileSize++;
	  }
  }
  // Rest Clusters Video
	for (i = 1; i < m_sfdFile.iVideoClusterCount-1; i++)
	{
		SetToPhysicalSector(hInFile, 0x100 + (lClusters[i] - 1) * 0x32000);
		for (j = 0; j < 100; j++)
		{
			ReadFile (hInFile, m_pBuffer, 0x100000, &dwRead, NULL);
			fOut.Write (m_pBuffer, 0x100000);
			m_pInfoVideo->m_nPosition++;
			m_pInfoVideo->m_nFileSize++;
		}
	}
  // Last Cluster Video
	SetToPhysicalSector(hInFile, 0x100 + (lClusters[m_sfdFile.iVideoClusterCount-1] - 1) * 0x32000);
	for (j = 0; j < (m_sfdFile.iVideoStopFirstCluster * 512) / 0x100000 ; j++)
	{
		ReadFile (hInFile, m_pBuffer, 0x100000, &dwRead, NULL);
		fOut.Write (m_pBuffer, 0x100000);
		m_pInfoVideo->m_nPosition++;
		m_pInfoVideo->m_nFileSize++;
	}
	ReadFile (hInFile, m_pBuffer, (m_sfdFile.iVideoStopFirstCluster * 512) % 0x100000, &dwRead, NULL);
	fOut.Write (m_pBuffer, (m_sfdFile.iVideoStopFirstCluster * 512) % 0x100000);

	fOut.Close ();
	tack = GetTickCount() - tick;
	m_pInfoVideo->m_nRipTime = (int)tack / 1000;
}

void CPVRFile::RestoreClusters()
{
	/*
	CString sFilename, sTemp;
	HANDLE hOutFile;
	CFile fIn;
	int i, j;
	long lClusters[MAX_FATPOINTERS];
	DWORD dwRead, tick=0,tack=0;

	// erst mal das Fileinfo wegschreiben
	sFilename = Filename ().c_str ();

	m_pInfoVideo->m_i64StartPos = 0;
	m_pInfoVideo->m_i64EndPos = 0;
	// Drive ffnen
	if (!OpenDrive(hOutFile, m_sDrivename))
	return;
	// erst die Audiocluster
	m_nRipType = RIP_TYPE_AUDIO;
	m_pInfoAudio->m_nEndPos = m_sfdFile.iAudioClusterCount * 100;
	m_pInfoAudio->m_nStartPos = 0;
	m_pInfoAudio->m_nPosition = 0;
	m_pInfoAudio->m_nFileSize = 0;
	m_pInfoAudio->m_bRunning = true;
	m_pInfoAudio->m_bFirstUpdate = true;

	sTemp.Format ("Restoring Audio Clusters (%d MB)", m_sfdFile.iAudioClusterCount * 100);
	Output (sTemp.GetBuffer (255));
	sTemp.ReleaseBuffer ();
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iAudioFatEntry);
	// Zeit beginn
	tick = GetTickCount();
	for (i = 0; i < m_sfdFile.iAudioClusterCount; i++)
	{
		sTemp.Format ("%s.audio_%02d", sFilename, i);
		if (!fIn.Open (sTemp, CFile::modeRead | CFile::typeBinary))
		{
			Output ("Audio Backup File could not be opened");
			tack = GetTickCount() - tick;
			m_pInfoAudio->m_nRipTime = (int)tack / 1000;
			return;
		}
		SetToPhysicalSector(hOutFile, 0x100 + (lClusters[i] - 1) * 0x32000);
		for (j = 0; j < 100; j++)
		{
			fIn.Read (m_pBuffer, 0x100000);
			WriteFile (hOutFile, m_pBuffer, 0x100000, &dwRead, NULL);
			m_pInfoAudio->m_nPosition++;
			m_pInfoAudio->m_nFileSize++;
		}
		fIn.Close ();
	}
	tack = GetTickCount() - tick;
	m_pInfoAudio->m_nRipTime = (int)tack / 1000;


	// jetzt die Videocluster
	m_nRipType = RIP_TYPE_VIDEO;
	m_pInfoVideo->m_nEndPos = m_sfdFile.iVideoClusterCount * 100;
	m_pInfoVideo->m_nStartPos = 0;
	m_pInfoVideo->m_nPosition = 0;
	m_pInfoVideo->m_nFileSize = 0;
	m_pInfoVideo->m_bRunning = true;
	m_pInfoVideo->m_bFirstUpdate = true;
	sTemp.Format ("Restoring Video Clusters (%d MB)", m_sfdFile.iVideoClusterCount * 100);
	Output (sTemp.GetBuffer (255));
	sTemp.ReleaseBuffer ();
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iVideoFatEntry);
	tick = GetTickCount();
	for (i = 0; i < m_sfdFile.iVideoClusterCount; i++)
	{
		sTemp.Format ("%s.Video_%02d", sFilename, i);
		if (!fIn.Open (sTemp, CFile::modeRead | CFile::typeBinary))
		{
			Output ("Video Backup File could not be opened");
			tack = GetTickCount() - tick;
			m_pInfoVideo->m_nRipTime = (int)tack / 1000;
			return;
		}
		SetToPhysicalSector(hOutFile, 0x100 + (lClusters[i] - 1) * 0x32000);
		for (j = 0; j < 100; j++)
		{
			fIn.Read (m_pBuffer, 0x100000);
			WriteFile (hOutFile, m_pBuffer, 0x100000, &dwRead, NULL);
			m_pInfoVideo->m_nPosition++;
			m_pInfoVideo->m_nFileSize++;
		}
		fIn.Close ();
	}
	tack = GetTickCount() - tick;
	m_pInfoVideo->m_nRipTime = (int)tack / 1000;
	CloseHandle (hOutFile);
	*/
}


void CPVRFile::Recover(int nNumber)
{
	long lClusters[MAX_FATPOINTERS];
	CString s;
	sprintf (m_sfdFile.cStationName, "Stations %d", nNumber);
	sprintf (m_sfdFile.cClipName, "Clip %d", nNumber);
	m_sfdFile.iAudioFatEntry = (short) (g_pfatFAT->GetEmergencyClusters (nNumber, 0xFE));
	m_sfdFile.iAudioClusterCount = (short) (g_pfatFAT->GetClusters (lClusters, m_sfdFile.iAudioFatEntry));
	m_sfdFile.iVideoFatEntry = (short) (g_pfatFAT->GetEmergencyClusters (nNumber, 0xFD));
	m_sfdFile.iVideoClusterCount = (short) (g_pfatFAT->GetClusters (lClusters, m_sfdFile.iVideoFatEntry));
	m_sfdFile.iLengthHours = 1;
	m_sfdFile.iLengthMinutes = 10;
	m_sfdFile.iLengthSeconds = 10;
	m_sfdFile.iRecordDay = 1;
	m_sfdFile.iRecordHour = 1;
	m_sfdFile.iRecordMinute = 1;
	m_sfdFile.iRecordMonth = 1;
	m_sfdFile.iRecordSecond = 0;
	m_sfdFile.iRecordYear = 2001;
	memset (m_sfdFile.ssiSkipIndex, 0, sizeof (m_sfdFile.ssiSkipIndex));
	WriteFileData (nNumber);
}

__int64 CPVRFile::GetNextPTS (unsigned int uiStreamID, CIBits* pcbStream, int nCluster,
                              int nStreamCluster, int & nOffset, HANDLE hFile, BYTE * pBuf)
{
 	CPES_Packet * pppPES;
  __int64 i64CurrentPTS(-1), i64Pos;
  DWORD dwRead;
  int i (0);
  Output ("CPVRFile::GetNextPTS\n", false, true);  

  sprintf (m_sOutBuffer, "Searching Cluster %d / StreamCluster %d - Offset %d - ",nCluster,nStreamCluster,nOffset);
  Output (m_sOutBuffer, false, true);  

  i64Pos = (__int64) (0x100 + (nCluster - 1) * 0x32000 + (nOffset / 512));
	SetToPhysicalSector(hFile,i64Pos);
	ReadFile (hFile, pBuf, 0x100000, &dwRead, NULL);

  while (i < 0x100000 && i64CurrentPTS == -1)
  {
	  for (; i < 0x100000; i++)
	  {
		  if (pBuf[i] == 0 && pBuf[i+1] == 0 && pBuf[i+2] == 0x01 && pBuf[i+3] == (BYTE) uiStreamID)
			  break;
	  }
	  i64Pos = nStreamCluster;
	  i64Pos *= (__int64) BITS_PER_CLUSTER;
	  i64Pos += (__int64) (i * 8);
    i64Pos += (__int64) (nOffset * 8);
	  pcbStream->SeekTo (i64Pos);
	  pppPES = new CPES_Packet(pcbStream, uiStreamID);	
	  i64CurrentPTS = pppPES->PTS ();
    delete pppPES;
    if (i64CurrentPTS == -1) i += 4;
  }
  sprintf (m_sOutBuffer, "PTS found: 0x%07I64X\n", i64CurrentPTS);
  Output (m_sOutBuffer, false, true);  
  return i64CurrentPTS;
}

unsigned __int64 CPVRFile::GetMaxPos (int nMaxInCluster, int nClusterCount)
{
  unsigned __int64 i64CurrentPTS;
	i64CurrentPTS = (unsigned __int64) nClusterCount-1;
	i64CurrentPTS *= (unsigned __int64) BITS_PER_CLUSTER;
	i64CurrentPTS += (unsigned __int64) (nMaxInCluster * 512 * 8);
  return i64CurrentPTS;
}
__int64 CPVRFile::GetCurrentPos (void)
{
  return m_i64CurrPos;
}

bool CPVRFile::GetClusterPTS (void)
{
	DWORD			tick=0,tack=0;
	CIBits* cbVideoStream=NULL;
  int i, nClusterCount, nSkippedBuffers (0), nOffset (0);
  unsigned int StreamID;
  HANDLE hFile;
	long lClusters[MAX_FATPOINTERS];

  tick = GetTickCount();
	*pfsLog << "CPVRFile::GetClusterPTS (void)" << endl;
	if (!OpenDrive(hFile, m_sDrivename))
	return false;
  nClusterCount = m_sfdFile.iVideoClusterCount;
  if (m_pi64ClusterPositions == NULL)
    m_pi64ClusterPositions = new __int64 [nClusterCount];
	cbVideoStream=new CIBits(m_sDrivename,m_sfdFile.iVideoFatEntry);
	m_nSkippedVideoBuffers = m_sfdFile.iVideoStartFirstCluster;
  m_nLastClusterOffset = m_sfdFile.iVideoStopFirstCluster;
  GetStreamIDs(*cbVideoStream, StreamID, nSkippedBuffers);
	g_pfatFAT->GetClusters (lClusters, m_sfdFile.iVideoFatEntry);
  for (i = 0; i < nClusterCount; i++)
  {
    nOffset = 0;
    if (i == 0) nOffset += (m_nFirstClusterOffset * 512);
    m_pi64ClusterPositions[i] = GetNextPTS (StreamID, cbVideoStream, 
     lClusters[i], i,  nOffset, hFile, m_pBuffer);
  }
  delete cbVideoStream;
	tack = GetTickCount()-tick;
	sprintf (m_sOutBuffer, "... it took %d  seconds...\n", (int)tack / 1000);
  Output (m_sOutBuffer);
	*pfsLog << "CPVRFile::GetClusterPTS (void)--end" << endl;
  return true;
}
