#ifndef PVRFILE_H
#define PVRFILE_H
#pragma once

#include "general.h"
#include "ibits.h"
#include "obits.h"

#ifdef _WINPVR
#include "..\IDInfos.h"
#include "..\InfoBlock.h"
#endif

#define TOC_OFFSET_SECTORS			0x0A
#define MAX_SKIP_INDEX				0x0A

#define RIP_TYPE_AUDIO				0x00
#define RIP_TYPE_VIDEO				0x01

#define ID_189                0x00
#define ID_190                0x01

struct sSkipIndex
{
	__int16		iStart;
	__int16		iStop;
};

struct sFileDescription_old
{
	__int8		iRecordNo;				
	char  		cStationName[0xD];
  char      cFWMarker;
	char	  	cClipName[0x29];
	__int16		iRecordYear;
	__int8		iRecordMonth;
	__int8		iRecordDay;
	__int8		iRecordHour;
	__int8		iRecordMinute;
	__int8		iRecordSecond;
	char		  cUnknown1;
	__int8		iLengthHours;
	__int8		iLengthMinutes;
	__int8		iLengthSeconds;

	char		  cUnknown2_1[0x05];
	__int16		iVideoClusterCount;
	__int16		iAudioClusterCount;
  __int32   iVideoStopFirstCluster;
  __int32   iAudioStopFirstCluster;

	__int16		iVideoFatEntry;
	__int16		iAudioFatEntry;
	sSkipIndex	ssiSkipIndex[MAX_SKIP_INDEX];
  __int32         iVideoStartFirstCluster;
  __int32         iAudioStartFirstCluster;
	char		cRemaining[0x178];
};

struct sFileDescription
{
  __int8          iRecordNo;                              
  char            cStationName[0x11];
  char            cFWMarker;
  char            cClipName[0x29];
  __int16         iRecordYear;
  __int8          iRecordMonth;
  __int8          iRecordDay;
  __int8          iRecordHour;
  __int8          iRecordMinute;
  __int8          iRecordSecond;
  char            cUnknown1;
  __int8          iLengthHours;
  __int8          iLengthMinutes;
  __int8          iLengthSeconds;

  char            cUnknown2_1[0x05];
  __int16         iVideoClusterCount;
  __int16         iAudioClusterCount;
  __int32         iVideoStopFirstCluster;
  __int32         iAudioStopFirstCluster;
//  char            cUnknown2_2[0x08];

  __int16         iVideoFatEntry;
  __int16         iAudioFatEntry;
  sSkipIndex      ssiSkipIndex[MAX_SKIP_INDEX];
  __int32         iVideoStartFirstCluster;
  __int32         iAudioStartFirstCluster;
  char            cRemaining[0x174];
};

class CPVRFile
{
public:
	bool m_bEndOverrun;
	void LookupClusters ();
  __int64 GetCurrentPos (void);
	void Recover (int nNumber);
	bool SearchStreamIDs(CIBits &InStream, unsigned int &StreamID, int & nSkippedBuffers);
	bool SearchForStart (__int64 i64BeginPTS, CIBits* pcbStream, unsigned int uiStreamID
		, __int16 iFatEntry, __int16 iClusterCount);
  bool GetClusterPTS (void);
	void SetFilename (const char * pFilename, bool bTake=true);
#ifdef _WINPVR	
	bool WriteFileData (unsigned __int16 nFilenummer);
	sFileDescription * GetFileDescription (void);
	bool * m_pbCancel;
	unsigned int m_nFileNo;
	CInfoBlock * m_pInfoAudio, * m_pInfoVideo;
#endif
	void Dump(ostream* posOut,bool Wait);
	CPVRFile(const string &Drivename, unsigned int FileNo);
	CPVRFile (const string &Drivename, sFileDescription * psfdFile);
	~CPVRFile();

	void RipStreams(double iStartSeconds,double iSeconds, bool bWriteLog=true);
	void dump_file ();
//private

private:
  __int64 GetNextPTS (unsigned int uiStreamID, CIBits* pcbStream, int nCluster, int nStreamCluster, int & nOffset, HANDLE hFile, BYTE * pBuf);
	string				Filename();
	sFileDescription_old	m_sfdFile_old;
  sFileDescription m_sfdFile;
	bool m_bBeginOverrun;
  __int64 m_i64FirstPTS;
	bool m_bOverrun;
	void BackupClusters ();
	void RestoreClusters ();
	int m_nFirstClusterOffset;
	CIDInfos m_IDsAudio, m_IDsVideo;
	bool m_bStreamWasCut;
	int m_nSkippedAudioBuffers;
	int m_nSkippedVideoBuffers;
	char m_cFilename [255];
	int m_nRipType;
	void RipAudioStreamRaw(__int64 i64StartPTS,__int64 i64EndPTS, CIBits* pcbStream, unsigned int uiStreamID);
  void RipVideoStreamRaw (__int64 i64StartPTS,__int64 i64EndPTS, CIBits* pcbStream, unsigned int uiStreamID);

	void Output (char * s, bool bOutConsole = true, bool bOutFile = true);
	string				m_sDrivename;
	bool				m_RecordNoOkay;
	vector <int>		m_RecoveryBuffer;

	void				AC3Recovery(long PTSDiff, COBits* pcoOutput);
	void				CopyBytes(unsigned int BytesToCopy, CIBits* pcbStream, COBits* pcoOutput);
	bool				GetStreamIDs(CIBits &InStream, unsigned int &StreamID, int & nSkippedBuffers);
	bool				GetFirstPTS(CIBits* pcbStream, unsigned int uiStreamID, __int64 &i64PTS);
	void				MPEG_Audio_Recovery(long PTSDiff, COBits* pcoOutput);
	void				RipAudioStream(CIBits* pcbStream, 
									COBits* pcoOutput, 
									unsigned int uiStreamID, 
									__int64 i64StartPTS , 
									__int64 i64EndPTS);
	void				RipPrivateStream(CIBits* pcbStream, 
									COBits* pcoOutput, 
									unsigned int uiStreamID, 
									__int64 i64StartPTS , 
									__int64 i64EndPTS);
	void				RipVideoStream(CIBits* pcbStream, 
									COBits* pcoOutput, 
									unsigned int uiStreamID, 
									__int64 i64StartPTS , 
									__int64 i64EndPTS);
	char				m_sOutBuffer[500];

  // Art der Firmware
  int m_nFirmwareType;
  __int64 * m_pi64ClusterPositions;
public:
  // Offset im letzten cluster
  __int32 m_nLastClusterOffset;
private:
  // maximale Videoposition
  unsigned __int64 m_i64VideoMaxPos;
  unsigned __int64 m_i64AudioMaxPos;
  unsigned __int64 GetMaxPos (int nMaxInCluster, int nClusterCount);
  __int64 m_i64PTSAudio, m_i64PTSVideo;
  __int64 m_i64CurrPos;
  BYTE * m_pBuffer;
public:
  double m_dAspectRatio;
private:
  double m_dOffsetTime;
public:
  double m_dRipTime;
  double m_dLenInSeconds;
};

#endif